require 'set'
require 'pp'
require 'tsort'

# A hash that can be used to represent a graph and find strongly 
#  connected components (and other functionality in TSort class)
# Copied word-for-word from ruby documentation on TSort
#  (http://www.ruby-doc.org/stdlib-2.0/libdoc/tsort/rdoc/TSort.html)
class Hash  
  include TSort
  alias tsort_each_node each_key
  def tsort_each_child(node, &block)
    fetch(node).each(&block)
  end
end


class Graph
	
	# Constructor
	# Pass true into 'isDirected' if disire a directed graph
	def initialize(isDirected=false)
		@directed = isDirected ? true : false
		@graph = {}	# the graph: @graph[node_i] = {node1 => { edge1 => [relations1], edge2 => [relations2]}, node2 => ...
		@nodes = [] 	# list of all nodes in the graph
		#@INFINITY = 1 << 64
	end
	
	
	# Add edges to the graph.  Multiple edges between two nodes are allowed.
	# Any additional information about an edge between two nodes can be stored in 'edge_attributes'
	def add_edge(node1, node2, edge_attributes=nil) 
		if not @graph.has_key?(node1)
			@graph[node1] = {node2 => [edge_attributes]}		 
		elsif not @graph[node1].has_key?(node2)
			@graph[node1][node2] = [edge_attributes]
		else
			@graph[node1][node2] << edge_attributes     
		end
		
		if not @directed
			if not @graph.has_key?(node2)
				@graph[node2] = {node1 => [edge_attributes]}		 
			elsif not @graph[node2].has_key?(node1)
				@graph[node2][node1] = [edge_attributes]
			else
				@graph[node2][node1] << edge_attributes     
			end
		end
		
		@nodes << node1  if not @nodes.include? node1
			
		@nodes << node2  if not @nodes.include? node2
		
	end
	
	
	# Finds all triples of nodes such that there is an edge from node1
	#  to node2, from node2 to node3, and from node1 to node3.
	# Triangles returned as an array of arrays: 
	#  [[node1, node2, node3, edge_attribute1_2, edge2_3, edge1_3], [...]]
	# Note: there will be repeats if this is not a directed graph
	def find_triangles2
		triples = @nodes.permutation(3).to_a
		triangles = []  # to return
		
		# Find triangles
		triples.each do |n0, n1, n2|			
			edges_01 = @graph[n0][n1] if @graph[n0]
			next if edges_01.nil?

			edges_12 = @graph[n1][n2] if @graph[n1]
			next if edges_12.nil?
			
			edges_02 = @graph[n0][n2] if @graph[n0]
			next if edges_02.nil?
			
			edges_01.each do |e1|
				edges_12.each do |e2|
					edges_02.each do |e3|
						#print [n0, n1, n2, e1, e2, e3] , "\n"
						triangles << [n0, n1, n2, e1, e2, e3] 
					end
				end
			end
		end
		
		return triangles
	end
	
	# Finds all triples of nodes such that there is an edge from node1 to 
	#  node2, from node3 to node2, and from node1 to node3.
	# Returns an array of arrays like:
	#  [[node1, node2, node3, edge_attribute1_2, edge3_2, edge1_3], [...]]
	# Note: there will be repeats if this is not a directed graph
	def find_triangles1
		triples = @nodes.permutation(3).to_a
		triangles = []  # to return
			
		# Find triangles
		triples.each do |n0, n1, n2|		
			edges_01 = @graph[n0][n1] if @graph[n0]
			next if edges_01.nil?

			edges_21 = @graph[n2][n1] if @graph[n2]
			next if edges_21.nil?
			
			edges_02 = @graph[n0][n2] if @graph[n0]
			next if edges_02.nil?
			
			edges_01.each do |e1|
				edges_21.each do |e2|
					edges_02.each do |e3|
						#print [n0, n1, n2, e1, e2, e3] , "\n"
						triangles << [n0, n1, n2, e1, e2, e3] 
					end
				end
			end
		end
		
		return triangles
	end
	
	
	# Finds all transitive ngons, with greater than 3 nodes.
	#   i.e. a set of nodes 0,1,...,N such that edges exist from nodes i to i+1, and 
	#   an edge exists from node 0 to node N.
	# Only one ngon is returned per unique set of nodes.
	#   (the shortest ngon; the first found if multiple shortest ngons found)
	# Returns a hash mapping a node and edge attribute to an array like:
	#   [edge_attribute0_n, node0] => [[edge_attribute0_1, node1], [edge1_2, node2], [edge2_3, node3], [...]] 
	#   where edge_attribute0_1 is the edge between node1 and node0, etc
	#   and edge_attribute0_n is the "transitive" edge between node0 and (the last) nodeN
	# Note: this method was created assuming a directed graph with max one relation between nodes.
	#   Not sure what results would be if multiple relation between nodes are allowed or undirected graph.
	def find_transitive_ngons
		ngons = {}   # to return

		find_all_paths 

		@nodes.each do |node|
			next if @graph[node].nil?
			pairs = @graph[node].keys.permutation(2).to_a
			
			# looking for paths from node (then n1) to n2
			pairs.each do |n1, n2|  #for all pair of outgoing rels to n1 and n2
				paths = return_paths(n1, n2) 
				next if paths.nil?

				# Only return shortest ngon
				shortest_path= nil  # store shortest of the paths found
				paths.each do |p|
					next if p.length <=  2  
						# don't want triangles; just 4-gons and above
					next if p.flatten.include? node  
						# make sure 'node' not on path btw n1 and n2

					shortest_path = p  if shortest_path.nil? or (p.length < shortest_path.length)
				end

				next if shortest_path.nil?
				shortest_path = Marshal.load( Marshal.dump(shortest_path) ) #deep copy
				
				# get relations between n,n1  and n,n2
				first_edge = @graph[node][n1][0]
				transitive_edge = @graph[node][n2][0]
				
				# replace first element of shortest_path (which is just n1) with [rel(n,n1), n1]		 
				first_node = shortest_path[0]  #  which also = n1
				shortest_path[0] = [first_edge, first_node]  ##########EDITING REFERENC!!!!!!!
				
				# make key [rel(n,n2), n] and save path in hash
				ngons[ [transitive_edge, node] ] = shortest_path
				 # ngons{[edge0_n, node0]} => [[edge0_1, node1], [edge1_2, node2], [edge2_3, node3], ]


			end	#pairs-of-outging-relations	
		end #all-nodess
		
		return ngons
	end
	
	
	# Returns a hash that maps each node in the graph to it's level (depth)
	#  in the graph. 
	# First cycles are removed by collapsing strongly connected components
	#  into a single node. Then levels are assigned as follows: Nodes with
	#  no incoming edges are assigned level 0.  Remaining nodes are given 
	#  the level of one more than the maximum of its predecessor nodes' levels.	
	def assign_levels
		
		# Find strongly connected components (SCCs)
		scc_graph = {}	# graph with only single edge between nodes, no edge attributes, 
					#  and an entry for all nodes(unlike @graph); Used to find scc's
					#(single edges ok since multiple edges don't effect SCC calculation)
		@nodes.each do |n|
			if @graph.has_key? n
				scc_graph[n] = @graph[n].keys  
			else
				scc_graph[n] = []
			end
		end

		sccs = scc_graph.strongly_connected_components

		# Create names for SCC nodes
		scc_nodes = {}  # maps a node that is part of an SCC to it's SCC node name
		sccs.each_index do |i|
			sc = sccs[i]
			if sc.length > 1				
				sc.each {|s|  scc_nodes[ s ] = "__SCC_#{i}" }
			end
		end

		# Create the new graph with the SCCs collapsed into a single node 
		#  (using the name created above, stored in 'scc_nodes')
		collapsed_scc_graph = Graph.new true  # graph with scc's collapsed
		@graph.each_pair do |from_node, edge_hash|  # populate 'collapsed_scc_graph'
			from_node = scc_nodes[from_node] if scc_nodes.has_key? from_node  
				# check whether from-node is part of an scc
					
			edge_hash.each_key do |to_node|
				to_node = scc_nodes[to_node] if scc_nodes.has_key? to_node
				
				next if from_node == to_node
				
				collapsed_scc_graph.add_edge(from_node, to_node)
			end
		end

		# Find levels of nodes in collapsed graph
		levels_scc = collapsed_scc_graph.get_node_levels

		# Map SCC nodes (and levels) back to actual nodes
		levels = {}		# to return.  Maps a node to it's level(depth) in the graph
		@nodes.each do |n|  
			if levels_scc.has_key? n
				levels[n] = levels_scc[n]
			else
				levels[n] = levels_scc[ scc_nodes[n] ]
			end
		end
		
		return levels

		
=begin	begin
			#sorted = scc_graph.tsort
			#pp sorted
		rescue TSort::Cyclic => e 
			puts e.message 
		end
=end

	end
	
	
	# Returns a hash mapping each node in this graph to a level, where
	#  nodes with no incoming edges are at level 0, and remaining nodes
	#  are at level one more than the max of their predecessors' levels.
	# Assumes a DAG (directed, acyclic graph).  Infinite recursion will 
	#  result if called on a cyclic graph.
	def get_node_levels
		levels = {}		
		@nodes.each do |n|
			levels[n] = get_depth(n) 
		end
		return levels
	end
	
	
	# Finds all simple paths between all nodes
	# Use return_paths() to get the paths found
	def find_all_paths
		# This method essentially populates '@paths'
		@paths = {}	# Maps each node n to a hash.  This hash maps a node m to
					#  a set of paths that start at n and end at m
				
		# Find all simple paths between all nodes
		all_paths = {}
		@nodes.each do |n|
			@my_paths = Set.new
			find_all_simple_paths(n)		
			all_paths[n] = @my_paths
		end
		
		# Populate @paths to organize the paths found
		@nodes.each do |n|
			n_paths = {}  # maps a node, m, to a set of paths that start from n and end at m
		
			# populate 'n_paths'
			all_paths[n].each do |p| 
				next if p.length == 1
				last_node = p[-1][1] 
				
				if n_paths.has_key? last_node
					n_paths[ last_node ] << p
				else
					n_paths[ last_node ] = [ p ]
				end
			end
			@paths[n] = n_paths
		end	
		
=begin	# Print all paths	
		all_paths.each_key do |s|
			print "Source node: #{s} \n"
			all_paths[s].each do |p|
				print p, "\n"
			end
			puts
		end
=end
	end
	
	# After find_all_paths() is called, use this method to obtain paths
	#  between 'source_node' and 'end_node'
	# Paths are of the form: [source_node, [edge_attr1, node1], [edge_attr2, node2,...  ]]
	# If end_node is not given, all paths to all other nodes is given as a hash
	#  whose key is another node in the graph and value are an array of paths
	def return_paths(source_node, end_node=nil)
		raise "Node not found: #{source_node}" unless @nodes.include? source_node
		
		if end_node.nil?
			return @paths[source_node] 
		else
			raise "Node not found: #{end_node}" unless @nodes.include? end_node
			return @paths[source_node][end_node]
		end
	end
	

private

	# Finds all simple paths between 'source' node and all other nodes
	def find_all_simple_paths(source, current_path=[source])
		@my_paths << current_path  #store this path
	
		# keep going recursively until a node repeats on the path, or reach a dead-end
		return nil  if @graph[source].nil?  # dead-end
		
		@graph[source].each do |node, paths|
			#if this neighbor node isn't already on path, follow it's outgoing paths
			if not current_path.flatten.include? node # (if not repeated node)
				paths.each do |path|					
					find_all_simple_paths( node, current_path + [[path, node]] )
				end
			end
		end
	end
	
		
	# Helper function to 'get_node_levels()'
	# Assumes 'node' is a node in this graph. Assumes graph is a DAG.
	def get_depth(node)
		# currently implemented as a recursive function.
		# can be made more efficient by storing depths already found.
		
		# find predecessor nodes
		predecessors = []
		@graph.each_pair do |pnode, edge_hash|
			predecessors << pnode  if edge_hash.has_key? node
		end
		
		
		if predecessors.length == 0 #no incoming edges
			return 0
		else #depth is 1 more than max of predecessor depths
			pred_depths = []
			predecessors.each do |n| 
				pred_depths << get_depth(n)
			end
			return 1 + pred_depths.max
		end

	end
		
end


if __FILE__ == $0
	isDirected = false
	gr = Graph.new isDirected
	gr.add_edge("a","b",5)
	gr.add_edge("b","c",3)
	gr.add_edge("c","d",1)
	gr.add_edge("a","d",10)
	gr.add_edge("b","d",2)
	gr.add_edge("b","d",3) # multiple edges between classes
	gr.add_edge("f","g",1)
	#gr.add_edge("d","a",1) #adds cycle
=begin	
	puts "All simple paths from node a:"
	gr.find_all_paths
	pp gr.return_paths "a"
=end	
	print "\n\n"
	puts "Triangles: "
	pp gr.find_triangles
end