require_relative 'smt_ar_processor'
require_relative 'csv_hash_formatter'
require 'optparse'
require 'fileutils'  # for rm -rf directory 
 
 
# Returns OS-specific dev/null string
def blackhole
    nullfile = "/dev/null" 
    if RUBY_PLATFORM =~ /mswin|mingw/
        nullfile = "NUL"
    end
    return nullfile
end


if __FILE__ == $0 then
	# Variables
	options = {} 
	model_directory = nil
	start_directory = Dir.getwd 
	INFERRED_PROPS_DIR = "inferred_properties"
	
		     
	# Set up OptionParser
	optparser = OptionParser.new do|opts|
		opts.banner = "Usage: ruby #{$0} [options] path\\to\\models\\folder" +
			"\n  This program infers properties about Ruby on Rails data models." +
			"\n  It then verifies them using Z3 and outputs results in CSV format."
			
		# Define the options
		options[:projection_on] = nil
		opts.on( '-j', '--projection_on', 'Perform property-based projection during verification' ) do
			options[:projection_on] = true
		end	
		
		opts.on( '-h', 'Display this screen' ) do
			puts opts
			exit
		end
	end
	
	
	# Parse options off of command line
	optparser.parse!
	
	# Get models folder
	if ARGV.length == 1  #model dir should be the only thing left on the cmd line
		model_directory = ARGV[0]
	else
		puts "  Internal error: Property inferer invoked incorrectly."
		#"  Program invoked incorrectly.  Use '-h' option for usage information."
		exit
	end

	 # Process each model file in the models folder
	 begin
		 Dir.chdir(model_directory)
	 rescue SystemCallError => e
		print "Error navigating to directory: " + e.to_s + "\n"
		exit
	 end
	 
	 Dir.foreach(".") do |filename|
		if filename =~ /\.rb$/ and File.file?(filename) then
			# Get contents of model files
			file = File.open(filename, 'r')
			file_contents = ""
			file.each {|line| file_contents << line }
			file.close

			# Parse
			s_exp = RubyParser.new.parse file_contents			
			#print s_exp  # print the s-expression generated
			ActiveRecordProcessor.new.process(s_exp)
		end
	end
	Dir.chdir start_directory
	
	
	# Infer properties and output SMT specifications for each property
	ActiveRecordProcessor.property_inference( INFERRED_PROPS_DIR, options[:projection_on] )

	#Check properties using SMT solver, Z3
	csv = CSVHashFormatter.new  # for CSV outputting
	solver = options[:projection_on] ? "z3+proj" : "z3"
	
	[:deletePropagation, :orphanPrevention, :transitive].each do |prop_type|
		num_pass = 0
		num_fail = 0
		num_err = 0
		num_to = 0
		Dir.foreach(INFERRED_PROPS_DIR) do |filename| 			
			if filename.start_with? prop_type.to_s
				smtfile = File.join(INFERRED_PROPS_DIR, filename)
								
				# Get number of variables and clauses produced
				num_vars = nil
				num_clauses = nil
				stats = %x[ruby stats_smtlibspec.rb #{smtfile} ]
				      
				if stats =~ /TOTAL variables: (\d+)/
					num_vars = $1
				end 
				if stats =~ /TOTAL clauses: (\d+)/
					num_clauses = $1
				end
				      
				      
				# Run SMT spec through Z3 for verification
				z3output = nil				
				z3time = timer do
					z3_cmd = "z3.exe  -T:10 -smt2 #{smtfile} "
					z3_cmd << " 2> #{blackhole}" if not $DEBUG # TODO: log translation warnings to file; right now just ignoring\hiding
					z3output = %x[#{z3_cmd}]
				end

				# Get property name and repair
				prop_name = nil
				repair_string =nil
				File.readlines(smtfile).each do |line|
					if line =~ /;Property: (.*)/
						prop_name = $1
					end
					if line =~ /;Repair: (.*)/
						repair_string = $1
					end
				end


				# Determine verification result (sat, unsat, timeout, or error)
				verif_result = :error 
				if not z3output =~ /^unsat/ and z3output =~ /\(error/                            
					verif_result = :error
					num_err += 1
				elsif z3output.scan(/\(error/).size > 1
					verif_result = :error                    
					num_err +=1
				else  
					z3output.each_line do |line|
						if line =~ /^sat/
							verif_result = :sat
							num_fail += 1
							break
						elsif line =~ /^unsat/
							verif_result = :unsat
							num_pass += 1
							break
						elsif line =~ /^timeout/
							verif_result = :timeout
							num_to += 1
							break
						end
					end

				end
				
puts smtfile   if verif_result == :error 
puts z3output   if verif_result == :error 
exit   if verif_result == :error 

				#Print results
				result = verif_result				
				repair = nil
				if verif_result == :sat  # output pass/fail instead of sat/unsat
					result = :fail
					repair = repair_string  # only print repairs for failing props
				elsif verif_result == :unsat
					result = :pass
				end
				
				csv << {
					:property	=> prop_name,
					:solver	=> solver,
					:num_vars	=> num_vars,
					:num_clauses => num_clauses,
					:verif_time	=> z3time,					
					:verif_result=> result,
					:repair	=> repair
				}
								
			end # each-file
		end
=begin		
		print prop_type, ": "
		print "\nNum Total: ", num_pass + num_fail + num_to + num_err
		print "\nNum Timeout: ", num_to
		print "\nNum Failed: ", num_fail
		print "\nNum Errors: ", num_err
		puts
		puts
=end		
	end  # each-property-type

	# Output results as csv
	puts csv 
	
	# Delete INFERRED_PROPS_DIR (temp directory for smt specs of inferred properties)
	FileUtils.rm_rf(INFERRED_PROPS_DIR)
end
