# The ActiveRecord Processor for translation to SMT-LIB

require 'rubygems'
require 'ruby_parser'
require 'sexp_processor'
require 'active_support/inflector'
require_relative 'smt_translation_engine'


class ActiveRecordProcessor < SexpProcessor
	@@idgen = 0
	@@smt_spec = RailsToSMTTranslationEngine.new								 

	def initialize
		super
		@require_empty = false  #it's okay to not process entire parse tree  (just first class encountered)
	end

	# Prints the SMT-LIBv2 specification
	def self.translate_and_print_smtspec( outputfile, projection_classes, 
	  projection_relations, deleteobjtype, property_info, projection_on )  
		@@smt_spec.translate
		projclasses, projrelations = @@smt_spec.create_property( property_info, projection_on ) if property_info
		projection_classes, projection_relations = projclasses, projrelations if projection_on
		@@smt_spec.print_smt_spec(outputfile, deleteobjtype, projection_classes, projection_relations)
	end

	# Infers properties and outputs SMT specifications for each property in the 'output_directory'
	# Set 'projection_on' to true if projection should be performed during translation to SMT
	def self.property_inference( output_directory, projection_on )
		@@smt_spec.translate
		@@smt_spec.property_inference( output_directory, projection_on )
	end
	
	# Creates an entity-relation diagram and outputs it to 'filename' as .dot file.
	#  If 'output_as_png' is true, then the ERD is output as .png file.
	def self.create_erd( filename, output_as_png=false )
		@@smt_spec.translate
		@@smt_spec.create_erd( filename, output_as_png )
	end
	
	
	def process_class(exp)
		unless @class_called_already then  # only process the first class encountered
			@class_called_already = true
		
			classs = exp.shift
			class_name = exp.shift
			parentclass = exp.shift
			body = exp.shift
			
			# Get classname
			@classname = class_name.to_s

			# Process contents of s-expression
			process(parentclass)
			process(body)
			
			return s(classs, @classname, parentclass, body)
		end
		return exp
	end
	
	def process_const(exp)
		if not @class_called_already  # check that we're inside a class definition
			return exp
		end

		unless @const_called_already   # only process the first class encountered
			@const_called_already = true
			const = exp.shift
			parent_classname = exp.shift
			
			# Initialize @association_info and @thruassociation_info for this class
			@@smt_spec.association_info[@classname] = []
			@@smt_spec.thruassociation_info[@classname] = []
			
			# Record parent classname in @class_table
			@@smt_spec.class_table[@classname] = parent_classname.to_s
			
			return s(const, parent_classname)
		end
		return exp
	end
	
	def process_call(exp)
		if not @class_called_already
			return exp
		end
		
		call = exp.shift
		param2 = exp.shift
		declaration = exp.shift
		arglist = exp.shift
		
		rel_types = [:belongs_to, :has_one, :has_many, :has_and_belongs_to_many]
		
		# Process association declaration
		if rel_types.include?(declaration)
			# Process the first lit you come across to get the relation name (see process_lit method)
			@get_relname = true   
			process(arglist)	
			
			relclassname = nil
			classname = @classname
			entry_already_exists_in_poly_classmap = nil
			dependency = DependencyType::NONE
			
			# Get name of related class
			if @specified_classname  # :class_name option is set
				relclassname = @specified_classname
				@specified_classname = nil
			else
				relclassname =  @relname.singularize.camelize
			end
		
			# Process and save relation based on option(s) set
			if @thru_option_class  # :through option set
				if @source
					relclassname = @source.singularize.camelize
					@source = nil
				end
				
				thruassociation = ThruAssociationInfo.new(declaration, @relname, 
					relclassname, @thru_option_class.singularize.camelize)
				@@smt_spec.thruassociation_info[classname].push(thruassociation)
				# @thru_option_class = nil  ## is set to nil later; need to access value again
			end
			
			if @conditions_set
				@conditions_set = false
				orig_relclassname = relclassname.dup
				relclassname << "_" << @@idgen.to_s
				@@idgen += 1
				
				@@smt_spec.class_table[relclassname] = orig_relclassname
				@@smt_spec.conditional_classes << relclassname
			end
			
			
			if @dependent == :delete or @dependent == :delete_all
				dependency = DependencyType::DELETE
			elsif @dependent == :destroy
				dependency = DependencyType::DESTROY
			end
			@dependent = nil
						
						
			if @polymorphic_class	
				if @@smt_spec.polymorphic_classmap.has_key? @polymorphic_class
					@@smt_spec.polymorphic_classmap[@polymorphic_class].push(classname)
					entry_already_exists_in_poly_classmap = true
				else
					@@smt_spec.polymorphic_classmap[@polymorphic_class] = [ classname ]
					classname = @polymorphic_class
					@@smt_spec.association_info[classname] = []
				end
				
				@polymorphic_class = nil
			end
			
			
			if (not @thru_option_class) and (not entry_already_exists_in_poly_classmap)
				association = AssociationInfo.new(declaration, @relname, relclassname, 
					dependency)
				@@smt_spec.association_info[classname].push(association)
			end
			@thru_option_class = nil
			
		end #association-declaration-processing
		
		return s(call, param2, declaration, arglist)
	end
	
	def process_lit(exp)
		if not @class_called_already
			return exp
		end
		
		lit = exp.shift
		name = exp.shift
	
		# Turn on flags to signal which options have been set on an association declaration
		if @get_relname
			@relname = name.to_s  # save varname in @relname for process_call() to use
			@get_relname = false
		elsif name == :class_name	#used to specify classname manually		
			@get_classname = true
		elsif name == :through
			@get_thru_class = true
		elsif @get_thru_class
			@thru_option_class = name.to_s
			@get_thru_class = false
		elsif name == :source  #used to specify through class name manually
			@get_source = true
		elsif @get_source
			@source = name.to_s
			@get_source = false
		elsif name == :conditions
			@conditions_set = true
		elsif name == :as
			@polymorphic = true
		elsif @polymorphic
			@polymorphic_class = name.to_s.capitalize
			@polymorphic = false
		elsif @dependent_set
			@dependent = name
			@dependent_set = false
		elsif name == :dependent
			@dependent_set = true
		end
		
	
		return s(lit, name)
	end
	
	def process_str(exp)
		if not @class_called_already
			return exp
		end
		
		key = exp.shift
		str = exp.shift
		
		# For when :class_name option is set (i.e. related class is specifically specified
		#  so it does not have to be inferred from relation name)
		if @get_classname			
			@get_classname = false
			@specified_classname = str
		end
		
		return s(key, str)
	end


end # end-class MyProcessor