#require 'rubygems'
require 'pp'	     	# for pretty printing 
require 'fileutils' 	# for FileUtils.rm_rf method
gem 'ruby-graphviz'
require 'graphviz'   	# for creating graphical representation of schema using GraphViz  
require_relative 'graph'
		

# Times how long a block of code takes to run
def timer  
	start = Time.now  
	yield  
	return Time.now - start
end

# Structs
AssociationInfo = Struct.new(:declaration_type, :relation_name, :relation_with, :dependency)
	#Note: :dependency is of type DependencyType
ThruAssociationInfo = Struct.new(:declaration_type, :relation_name, :relation_with, 
	:thru_relation_class)

class DependencyType
	NONE = 1
	DELETE = 2
	DESTROY = 3
end 

class RailsToSMTTranslationEngine
	attr_accessor :class_table, :association_info, :thruassociation_info, :polymorphic_classmap,
		:conditional_classes
	
	Relation = Struct.new(:relation_name, :classB_name, :classA_name, 
		:relation_type)    # Note: :relation_type is of type RelationType
	ThruRelation = Struct.new(:relation_name, :thruclass_name, :relation1,
		:relation2, :thru_reltype)  # Note: :thru_reltype is of type ThroughRelType,
		# :relation1 and :relation2 are of type Relation
	DependencyInfo = Struct.new(:relation_name, :relation_with, :argument, :dependency_type)
		#Note: :argument is a boolean; true if this DependencyInfo's class is the argument in 
		#  the relation and false otherwise
		#:dependency_type is of type DependencyType
	
	def initialize
		@class_table = {}         # a hash that maps class names to parent class names
		@conditional_classes = []  # list of conditional subclasses created
		@association_info = {}   # a list of AssociationInfo elements, hashed using classname
		@thruassociation_info = {} # a list of ThruAssociationInfo elements, hashed using classname
		@polymorphic_classmap = {} # maps a superclass to a set of polymorphically-related classes 
		
		@polymorphic_parentsmap = {} #maps a polymorphic class to all its superclasses 
		                                            #  (inverse of @polymorphic_classmap)
		@top_level_sorts = []	   # a list of class names whose parent is ActiveRecord
		@subset_table = {}       # a hash that maps classes to their parent class name (i.e. for classes
						  #  whose base class isnt ActiveRecord and for conditional subclasses)
		@relations = []              # list of all relations (list of Relation objects) (including :through relations)
		@through_relations = []  # list of relations with :through option set (list of ThruRelation objects)
		
		@through_mapping = {} # maps ThruAssociationInfo objects to ThruRelation objects
		@dependent_rels = {}    # a hash that maps classes to a list of relation names that had 
		                                   #  the :dependent option set on their side
		@belongsto_rels = {}     # a hash that maps classes to a list of relation names that had 
		                                   #  the :belongs_to declaration on their side
		@property = nil
		@RAILS_HELPER_CLASSES = ["ActionMailer", "ActionPack", "ActionWebService", 
				"ActiveSupport", "Authlogic"]
		@translated = false		# true if the translate() method has been run
	end		
	

	# Assumes public variables have been populated
	# Populates private variables 
	# Returns true if successful
	def translate() 
		belongs_to_assocs = []  # to store belongs_to associations whose match has been found;
						 #  in case another relation is using this association as a match as 
						 #  well (e.g. multiple conditional relations can match with a single 
						 #  belongs_to declaration)
		
		# Initialize @polymorphic_parentsmap as inverse of @polymorphic_classmap
		@polymorphic_classmap.each_key do |key| 
			@polymorphic_classmap[key].each do |val|
				if @polymorphic_parentsmap.has_key? val
					@polymorphic_parentsmap[val].push key
				else
					@polymorphic_parentsmap[val] = [key]
				end
			end
		end

		# Match up relations in @association_info and put them in @relations
		associations = Marshal.load(Marshal.dump(@association_info))  #deep copy
		classes = associations.keys
		
		# Initialize hashes
		classes.each do |klass|
			@dependent_rels[klass] = []
			@belongsto_rels[klass] = []
		end
		
		classes.each do |klass|	
			associations[klass].each do |association|				
				relklass = association[:relation_with]
				parentclass = @class_table[relklass]
				if not associations.has_key? relklass and not associations.has_key? parentclass 
					$stderr.print ";warning 1: matching declaration not found (for " +
						association[:relation_name] +" relation in " + klass + ")\n"
				else
					matchingassoc = nil
					subset_relation = false
					
					if associations.has_key? relklass
						matchingassoc = associations[relklass].find {|a| a[:relation_with] == klass}
					end
					
					if not matchingassoc # then maybe this was stored as a relation with my superclass, 
						                     #   as would happen if :conditions option is set
						if associations.has_key? parentclass						
							matchingassoc = 
								@association_info[parentclass].find {|a| a[:relation_with] == klass}
							subset_relation = true
						end
					end
					
					# NEEDS TO BE TESTED:
					if not matchingassoc # then maybe this was stored as a relation with relclass's superclass, 
						                     #   as would happen if :conditions option is set
						if associations.has_key? relklass					
							matchingassoc = 
								@association_info[relklass].find {|a| @class_table[a[:relation_with]] == klass}
							#subset_relation = true  ##need?
							#klass = matchingassoc[:relation_with] if matchingassoc
							next if matchingassoc  #let relclass deal with creating this relation instead of reassigning 'klass' like above
						end
					end
					#TEST: class A has_many classB.  conditional on has_many side.  and then on belongs_to side
					#  Then multiple has_many conditionals.  Then multiple belongs_to conditionals.
					#  Then put the belongs_to declaration in classA and repeat.
					#  Then switch names of classA and classB and repeat. (testing order that declarations are processed)
					
					if not matchingassoc # then maybe the matching belongs_to was taken by another relation
						matchingassoc = belongs_to_assocs.find {|a| a[:relation_with] == klass}
					end
					
					# If match still not found and this is a belongs_to declaration, create a matching
					#  association myself  (app. developer didn't include the matching declaration, but
					#  semantically this is a 1-to-many relation)
					if not matchingassoc and association[:declaration_type] == :belongs_to
						matchingassoc = AssociationInfo.new(:has_many, 
							klass.downcase.pluralize, klass, DependencyType::NONE)
					end
					
					if not matchingassoc
						$stderr.print ";warning 2: matching declaration not found (" + association[:relation_name] 
						$stderr.print " relation in #{klass} with #{relklass})\n"
					else
						reltype1 = association[:declaration_type]
						reltype2 = matchingassoc[:declaration_type]
						relname1 = association[:relation_name]
						relname2 = matchingassoc[:relation_name]
						dependent1 = association[:dependency]
						dependent2 = matchingassoc[:dependency]
						relation = nil
						belongsto_class = nil		
						dependancy_info = nil

						if reltype1 == :has_and_belongs_to_many and reltype2 == :has_and_belongs_to_many
							relation = Relation.new(relname1 + "_" + relname2, 
								relklass, klass, RelationType::MANY_TO_MANY)
						elsif reltype1 == :belongs_to and reltype2 == :has_many
							relation = Relation.new(relname1 + "_" + relname2, 
								relklass, klass, RelationType::ONE_TO_MANY)
							belongsto_class = klass
						elsif reltype1 == :belongs_to and reltype2 == :has_one
							relation = Relation.new(relname1 + "_" + relname2, 
								relklass, klass, RelationType::ONE_TO_ONE)
							belongsto_class = klass
						elsif reltype1 == :has_many and reltype2 == :belongs_to
							relation = Relation.new(relname2 + "_" + relname1, 
								klass, relklass, RelationType::ONE_TO_MANY)
							belongsto_class = relklass
						elsif reltype1 == :has_one and reltype2 == :belongs_to
							relation = Relation.new(relname2 + "_" + relname1, 
								klass, relklass, RelationType::ONE_TO_ONE)
							belongsto_class = relklass
						else
							$stderr.print ";warning 3: unexpected declaration pair\n"
						end
						
						if relation
							@relations.push(relation)
							
							if not subset_relation
								associations[relklass].delete matchingassoc
							else
								belongs_to_assocs << matchingassoc
								associations[parentclass].delete matchingassoc
							# Don't delete association entry if :conditions option set because
							#   other relations in klass may be connecting with this relation 
							#   declaration in parentclass 
							end
						
							# Store information to handle :dependent option
							if belongsto_class
								if @belongsto_rels.include? belongsto_class
									@belongsto_rels[belongsto_class].push(
										[relation[:classB_name], relation[:relation_name]] )
								else
									@belongsto_rels[parentclass].push(
										[relation[:classB_name], relation[:relation_name]] )
								end
							end
							if dependent1 != DependencyType::NONE
								dependancy_info = DependencyInfo.new( relation[:relation_name],
									relklass, reltype1 == :belongs_to, dependent1)
								
								@dependent_rels[klass].push dependancy_info
							end
							if dependent2 != DependencyType::NONE
								dependancy_info = DependencyInfo.new( relation[:relation_name],
									klass, reltype2 == :belongs_to, dependent2)
								if @dependent_rels.include? relklass
									@dependent_rels[relklass].push dependancy_info
								else
									@dependent_rels[parentclass].push dependancy_info
								end
							end						
						end	
						
					end
				end
			end
			associations[klass] = []
		end #end relation matching		
		
		# Create entries in @relations and @through_relations based on entries in @thruassociation_info
		@thruassociation_info.each do |klass, associationlist|
			associationlist.each do |association|
				reltype = association[:declaration_type]
				relname = association[:relation_name]
				relclass = association[:relation_with]
				thru_relclass = association[:thru_relation_class]
				
				# TODO: Cases where both :through and :conditions option are set (maybe on opposite ends)
				# parentclass = @class_table[relklass]
				
				relation = nil
				thru_relname =  klass.downcase + "_" + relname
				
				case reltype 
					when :has_many
						relation = Relation.new(thru_relname, klass, 
							relclass, RelationType::ONE_TO_MANY)
					when :has_one
						relation = Relation.new(thru_relname, klass, 
							relclass, RelationType::ONE_TO_ONE)
					else   
						$stderr.print ";warning 4: unexpected error when translating ", relname, " relation in ", 
							klass, " -- :through option not set on has_many or has_one declaration\n"
				end 
				
				if relation					
					relation1 = @relations.find {|x| x[:classB_name] == klass and 
						x[:classA_name] == thru_relclass and 
						x[:relation_type] != RelationType::MANY_TO_MANY}
					relation2 = @relations.find {|x| x[:classB_name] == relclass and 
						x[:classA_name] == thru_relclass and 
						x[:relation_type] != RelationType::MANY_TO_MANY}
					thrureltype = ThroughRelType::TYPE2

					if not relation2
						relation2 = @relations.find {|x| x[:classA_name] == relclass and 
							x[:classB_name] == thru_relclass and 
							x[:relation_type] != RelationType::MANY_TO_MANY}
						thrureltype = ThroughRelType::TYPE1
					end
					
					if not relation1 or not relation2
						$stderr.print ";warning 5: unable to set up transitive relation from ", klass, " to ", relclass, "\n"
						#print @relations
=begin						
						$stderr.print "; ", thru_relclass, "\n"
						if not relation1  #TODO: are there other possibities for rel1 and rel2 (i.e. classA and B switched)
							$stderr.puts ";ERRROORRRRR! didn't find relation1 correctly!!!"
						end
						if not relation2
							$stderr.puts ";ERRROORRRRR! didn't find relation2 correctly!!!"
						end
=end						
						next
					end
					
					thru_relation = ThruRelation.new(thru_relname, thru_relclass,
						relation1, relation2, thrureltype)
					@through_relations.push thru_relation
					@through_mapping[association] = thru_relation
					
					relation_exists = @relations.find {|x| x[:relation_name] == relation[:relation_name]}
						# usually doesn't exist, but may. such as in a using :through option to 
						#  create a join table (i.e. a double hmthru relation between classes)
					@relations.push(relation) if not relation_exists
				end
			end
		end #creating relations from Through Association Info table
		
		# Identify top-level sorts and classes containing polymorphic relations
		polymorphic_classes = @polymorphic_parentsmap.keys

		@class_table.each_pair do |klass, parentclass|
			if not polymorphic_classes.include? klass
				if parentclass == "ActiveRecord"
					@top_level_sorts.push klass
				elsif not @RAILS_HELPER_CLASSES.include? parentclass
					@subset_table[klass] = parentclass
				end
			end
		end
		
		@translated = true

	end #end-fxn translate()
	
		
	
	# Creates a property using the information passed in
	# Populates @property upon successful creation
	#  'property_info' is of the form [property_name, arg1, arg2[, arg3] ]
	#  'projection_on' is true if projection classes and relations should be returned
	def create_property( property_info, projection_on )
		translate() if not @translated

		proj_classes = nil
		proj_relations = nil

		if property_info[0] == "alwaysRelated"
			
			classA = property_info[1]
			classB = property_info[2]
			switched = false
			relation = @relations.find {|x| (x[:classA_name] == classA or @subset_table[x[:classA_name]] == classA) and
					(x[:classB_name] == classB or @subset_table[x[:classB_name]] == classB or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classB)) ) and
					x[:relation_type] != RelationType::MANY_TO_MANY}
			if relation.nil?
				switched = true
				relation = @relations.find {|x| (x[:classA_name] == classB or @subset_table[x[:classA_name]] == classB) and
					(x[:classB_name] == classA or @subset_table[x[:classB_name]] == classA or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classA)) ) and
					x[:relation_type] != RelationType::MANY_TO_MANY}
			end
			if relation.nil?
				puts "Error: Could not find a relation between specified classes " +
					"(#{classA} and #{classB})."
				return
			end
			
			relname = relation[:relation_name]
			isPolyA = @polymorphic_parentsmap.has_key? classA
			isPolyB = @polymorphic_parentsmap.has_key? classB
			related = switched ? "(= a (#{relname} b))" : "(= b (#{relname} a))"
			
			if not isPolyA and not isPolyB
				@property = "(assert (not (forall ((a #{classA})) (exists ((b #{classB}))" +
						  "\n  #{related}" + "\n))))"
			elsif isPolyA and not isPolyB
				@property = "(assert (not (forall ((a PolymorphicClass)) " + 
						  "(exists ((b #{classB}))" +
						  "\n  (=> (is#{classA} a) #{related}" + " )\n))))"
			elsif not isPolyA and isPolyB
				@property = "(assert (not (forall ((a #{classA})) " +
						  "(exists ((b PolymorphicClass))" + 
						  "\n  (and (is#{classB} b) #{related}" + " )\n))))"
			else
				@property = "(assert (not (forall ((a PolymorphicClass)) " + 
						  "(exists ((b PolymorphicClass))\n  (=> (is#{classA} a) " +
						  "(and (is#{classB} b) #{related}" + " ))\n))))"
			end
			
			proj_classes = [classA, classB]
			proj_relations = [relname]


		elsif property_info[0] == "someUnrelated"
			classA = property_info[1]
			classB = property_info[2]
			switched = false				  
			relation = @relations.find {|x| (x[:classA_name] == classA or @subset_table[x[:classA_name]] == classA) and
					(x[:classB_name] == classB or @subset_table[x[:classB_name]] == classB or 
					(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classB)) ) }
			if relation.nil?
				switched = true
				relation = @relations.find {|x| (x[:classA_name] == classB or @subset_table[x[:classA_name]] == classB) and
					(x[:classB_name] == classA or @subset_table[x[:classB_name]] == classA or 
					 (@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classA)) ) }
			end
				
			if relation.nil?
				puts "Error: Could not find a relation between specified classes " +
					"(#{classB} and #{classA})."
					#puts
					#pp @relations
				return
			end
			
			relname = relation[:relation_name]
			isMany2Many = (relation[:relation_type] == RelationType::MANY_TO_MANY)
			isPolyA = @polymorphic_parentsmap.has_key? classA
			isPolyB = @polymorphic_parentsmap.has_key? classB
			related = nil
			
			if not isMany2Many and not switched 
				related = "(= b (#{relname} a))"
			elsif not isMany2Many and switched 
				related = "(= a (#{relname} b))"
			elsif isMany2Many and not switched 
				related = "(#{relname} a b)"
			else
				related = "(#{relname} b a)"
			end
			
			
			if not isPolyA and not isPolyB
				@property = "(assert (exists ((a #{classA})) (forall ((b #{classB}))" +
						  "\n  (not #{related})" + "\n)))"
			elsif isPolyA and not isPolyB
				@property = "(assert (exists ((a PolymorphicClass)) (forall ((b #{classB}))" +
						  "\n  (and (is#{classA} a)  (not #{related}) )" + "\n)))"
			elsif not isPolyA and isPolyB
				@property = "(assert (exists ((a #{classA})) (forall ((b PolymorphicClass))" +
						  "\n  (=> (is#{classB} b)  (not #{related}) )" + "\n)))"
			else
				@property = "(assert (exists ((a PolymorphicClass)) (forall ((b PolymorphicClass))" +
						  "\n  (and (is#{classA} a) (=> (is#{classB} b) (not #{related}) ))" + 
						  "\n)))"
			end
			
			proj_classes = [classA, classB]
			proj_relations = [relname]


		elsif property_info[0] == "multipleRelated"
		
			classA = property_info[1]
			classB = property_info[2]
			switched = false				  
			relation = @relations.find {|x| (x[:classA_name] == classA or @subset_table[x[:classA_name]] == classA) and
					(x[:classB_name] == classB or @subset_table[x[:classB_name]] == classB or 
					 (@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classB)) ) }
			if relation.nil?
				switched = true
				relation = @relations.find {|x| (x[:classA_name] == classB or @subset_table[x[:classA_name]] == classB) and
					(x[:classB_name] == classA or @subset_table[x[:classB_name]] == classA or 
					 (@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classA)) ) }
			end
				
			if relation.nil?
				puts "Error: Could not find a relation between specified classes " +
					"(#{classB} and #{classA})."
					#puts
					#pp @relations
				return
			end
			
			relname = relation[:relation_name]
			isMany2Many = (relation[:relation_type] == RelationType::MANY_TO_MANY)
			isPolyA = @polymorphic_parentsmap.has_key? classA
			isPolyB = @polymorphic_parentsmap.has_key? classB
			related1 = nil
			related2 = nil
			
			if not isMany2Many and not switched 
				related1 = "(= b1 (#{relname} a))"
				related2 = "(= b2 (#{relname} a))"
			elsif not isMany2Many and switched 
				related1 = "(= a (#{relname} b1))"
				related2 = "(= a (#{relname} b1))"
			elsif isMany2Many and not switched 
				related1 = "(#{relname} a b1)"
				related2 = "(#{relname} a b2)"
			else
				related1 = "(#{relname} b1 a)"
				related2 = "(#{relname} b2 a)"
			end
			
			
			if not isPolyA and not isPolyB
				@property = "(assert (exists ((a #{classA})(b1 #{classB})(b2 #{classB}))" +
						  "\n  (and (not (= b1 b2)) #{related1} #{related2} )"   + "\n))"
			elsif isPolyA and not isPolyB
				@property = "(assert (exists ((a PolymorphicClass)(b1 #{classB})(b2 #{classB}))" +
						  "\n  (and (is#{classA} a) (not (= b1 b2)) #{related1} #{related2} )" + 
						  "\n))"
			elsif not isPolyA and isPolyB
				@property = "(assert (exists ((a #{classA})(b1 PolymorphicClass)(b2 PolymorphicClass))" +
						  "\n  (and (is#{classB} b1) (is#{classB} b2) " + 
						  "(not (= b1 b2)) #{related1} #{related2} )" +  "\n))"
			else
				@property = "(assert (exists ((a PolymorphicClass)(b1 PolymorphicClass)(b2 PolymorphicClass))" +
						  "\n  (and (is#{classA} a) (is#{classB} b1) (is#{classB} b2) " + 
						  "(not (= b1 b2)) #{related1} #{related2} )" +  "\n))"
			end
			
			proj_classes = [classA, classB]
			proj_relations = [relname]
			

		elsif property_info[0] == "transitive"
					
			classA = property_info[1]
			classB = property_info[2]
			classC = property_info[3]
			proj_classes = [classA, classB, classC]
			isPolyA = @polymorphic_parentsmap.has_key? classA
			isPolyB = @polymorphic_parentsmap.has_key? classB
			isPolyC = @polymorphic_parentsmap.has_key? classC
			relname1 = nil
			relname2 = nil
			relname3 = nil
			switched1 = false
			switched2 = false
			switched3 = false
			thruassociation = @thruassociation_info[classA].find {
				|x| x[:relation_with] == classC and x[:thru_relation_class] == classB }
			
			if not thruassociation.nil?
				thrurelation = @through_mapping[thruassociation]
				relname1 = thrurelation[:relation1][:relation_name]
				relname2 = thrurelation[:relation2][:relation_name]
				relname3 = thrurelation[:relation_name]
				
				if thrurelation[:thru_reltype] == ThroughRelType::TYPE2
					switched1 = true
				end

=begin
				if thrurelation[:thru_reltype] == ThroughRelType::TYPE1
					if not isPolyA
						@property = "(assert (not (forall ((a #{classA}))\n\t" +
							"(= (" + relname + " a) (" + relation1 + 
							" (" + relation2 + " a)) )\n)))\n" 					
					else
						@property = "(assert (not (forall ((a PolymorphicClass))\n\t" +
							"(=> (is#{classA} a) (= (" + relname + " a) (" + relation1 + 
							" (" + relation2 + " a)) )\n))))\n" 					
					end
				else
					if not isPolyB
						@property = "(assert (not (forall ((b #{classB}))\n\t" + 
							"(= (" + relation1 + " b) (" + relname + 
							" (" + relation2 + " b)) )\n)))\n"
					else
						@property = "(assert (not (forall ((b PolymorphicClass))\n\t" + 
							"(=> (is#{classB} b) (= (" + relation1 + " b) (" + 
							relname + " (" + relation2 + " b)) )\n))))\n"
					end
				end
			
				proj_classes = [classA, classB, classC]
				proj_relations = [relname, relation1, relation2]
=end					
			else
				relation1 = @relations.find {|x| (x[:classA_name] == classA or @subset_table[x[:classA_name]] == classA) and
						(x[:classB_name] == classB or @subset_table[x[:classB_name]] == classB or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classB)) ) and
						x[:relation_type] != RelationType::MANY_TO_MANY}
				if relation1.nil?
					switched1 = true
					relation1 = @relations.find {|x| (x[:classA_name] == classB or @subset_table[x[:classA_name]] == classB) and
						(x[:classB_name] == classA or @subset_table[x[:classB_name]] == classA or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classA)) ) and
						x[:relation_type] != RelationType::MANY_TO_MANY}
				end
				relation2 = @relations.find {|x| (x[:classA_name] == classB or @subset_table[x[:classA_name]] == classB) and
						(x[:classB_name] == classC or @subset_table[x[:classB_name]] == classC or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classC)) ) and
						x[:relation_type] != RelationType::MANY_TO_MANY}
				if relation2.nil?
					switched2 = true
					relation2 = @relations.find {|x| (x[:classA_name] == classC or @subset_table[x[:classA_name]] == classC) and
						(x[:classB_name] == classB or @subset_table[x[:classB_name]] == classB or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classB)) ) and
						x[:relation_type] != RelationType::MANY_TO_MANY}
				end		
				relation3 = @relations.find {|x| (x[:classA_name] == classC or @subset_table[x[:classA_name]] == classC) and
						(x[:classB_name] == classA or @subset_table[x[:classB_name]] == classA or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classA)) ) and
						x[:relation_type] != RelationType::MANY_TO_MANY}
				if relation3.nil?
					switched3 = true
					relation3 = @relations.find {|x| (x[:classA_name] == classA or @subset_table[x[:classA_name]] == classA) and
						(x[:classB_name] == classC or @subset_table[x[:classB_name]] == classC or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classC)) ) and
						x[:relation_type] != RelationType::MANY_TO_MANY}
				end	
					
				if not relation1 or not relation2 or not relation3
					puts "Error: Could not find relations between specified classes " +
						"(#{classA}, #{classB}, and #{classC})."
					return
				end
				
				relname1 = relation1[:relation_name]
				relname2 = relation2[:relation_name]
				relname3 = relation3[:relation_name]
			end
			
			polyLineB = isPolyB ? "(is#{classB} b)" : nil
			polyLineAC = nil
			if isPolyA and isPolyC
				polyLineAC = "(and (is#{classA} a) (is#{classC} c) )"
			elsif isPolyA
					polyLineAC = "(is#{classA} a)"
			elsif isPolyC
					polyLineAC = "(is#{classC} c)"
			end
			
			related1 = switched1 ? "(= a (#{relname1} b))" : 
							"(= b (#{relname1} a))"
			related2 = switched2 ? "(= b (#{relname2} c))" : 
							"(= c (#{relname2} b))"
			related3 = switched3 ? "(= c (#{relname3} a))" :
							"(= a (#{relname3} c))" 
			classA = "PolymorphicClass" if isPolyA
			classB = "PolymorphicClass" if isPolyB
			classC = "PolymorphicClass" if isPolyC

			@property = "(assert (not (forall ((a #{classA})(c #{classC}))"
			@property << "\n  (=> #{polyLineAC}" if polyLineAC
			@property << "\n    (and "
			@property << "\n    (=> #{related3} "
			@property << "\n        (exists ((b #{classB})) (and #{related1} #{related2} #{polyLineB})))"
			@property << "\n    (=> (exists ((b #{classB})) (and #{related1} #{related2} #{polyLineB}))"
			@property << "\n        #{related3})"
			@property << "\n     )"
			@property << "\n  )" if polyLineAC
			@property << "\n)))"			
			
			proj_relations = [relname1, relname2, relname3]


		elsif property_info[0] == "noOrphans"
			
			classA = property_info[1]
			classB = property_info[2]
			proj_classes = [classA, classB]
			isPolyA = @polymorphic_parentsmap.has_key? classA
			isPolyB = @polymorphic_parentsmap.has_key? classB
			switched = false
			
			relation = @relations.find {|x| (x[:classA_name] == classA or @subset_table[x[:classA_name]] == classA) and
					(x[:classB_name] == classB or @subset_table[x[:classB_name]] == classB or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classB)) ) }
			if relation.nil?
				switched = true
				relation = @relations.find {|x| (x[:classA_name] == classB or @subset_table[x[:classA_name]] == classB) and
					(x[:classB_name] == classA or @subset_table[x[:classB_name]] == classA or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classA)) ) }
			end
			if relation.nil?
				puts "Error: Could not find a relation between specified classes " +
					"(#{classA} and #{classB})."
				return
			end
			
			relname = relation[:relation_name]
			related = switched ? "(Post_#{relname} b a)" :  "(Post_#{relname} a b)"
			postA = "(Post_#{classA} a)"
			postB = "(Post_#{classB} b)"
			polyLineA = isPolyA ? "(is#{classA} a)" : nil
			polyLineB = isPolyB ? "(is#{classB} b)" : nil
			classA = "PolymorphicClass" if isPolyA
			classB = "PolymorphicClass" if isPolyB
			
			@property = "(forall ((b #{classB})) \n"
			@property << "(=> #{postB} (exists ((a #{classA}))"
			@property << "\n  (=> #{polyLineB}" if polyLineB
			@property << "\n    (and #{related} #{postA} #{polyLineA})"
			@property << "\n  )" if polyLineB
			@property << "\n)))"	
			
			proj_relations = [relname]

		
		elsif property_info[0] == "noDangling"
			
			classA = property_info[1]
			classB = property_info[2]
			proj_classes = [classA, classB]
			isPolyA = @polymorphic_parentsmap.has_key? classA
			isPolyB = @polymorphic_parentsmap.has_key? classB
			switched = false
			relation = @relations.find {|x| (x[:classA_name] == classA or @subset_table[x[:classA_name]] == classA) and
					(x[:classB_name] == classB or @subset_table[x[:classB_name]] == classB or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classB)) ) }
			if relation.nil?
				switched = true
				relation = @relations.find {|x| (x[:classA_name] == classB or @subset_table[x[:classA_name]] == classB) and
					(x[:classB_name] == classA or @subset_table[x[:classB_name]] == classA or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classA)) ) }
			end
			if relation.nil?
				puts "Error: Could not find a relation between specified classes " +
					"(#{classA} and #{classB})."
				return
			end
			
			relname = relation[:relation_name]
			related = switched ? "(Post_#{relname} b a)" :  "(Post_#{relname} a b)"
			postA = "(Post_#{classA} a)"

			polyLine = nil
			if isPolyA and isPolyB
				polyLine = "(and (is#{classA} a) (is#{classB} b))"
			elsif isPolyB
				polyLine = "(is#{classB} b)"
			elsif isPolyA
				polyLine = "(is#{classA} a)"
			end
			
			classA = "PolymorphicClass" if isPolyA
			classB = "PolymorphicClass" if isPolyB
			
			@property = "(forall ((a #{classA})(b #{classB}))"
			@property << "\n  (=> #{polyLine}" if polyLine
			@property << "\n    (=> #{related} #{postA} )"
			@property << "\n  )" if polyLine
			@property << "\n)"	
			
			proj_relations = [relname]
			
		
		elsif property_info[0] == "deletePropagates"
			
			classA = property_info[1]
			classB = property_info[2]
			proj_classes = [classA, classB]
			isPolyB = @polymorphic_parentsmap.has_key? classB
			switched = false
			
			relation = @relations.find {|x| (x[:classA_name] == classA or @subset_table[x[:classA_name]] == classA) and
					(x[:classB_name] == classB or @subset_table[x[:classB_name]] == classB or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classB)) ) }
			if relation.nil?
				switched = true
				relation = @relations.find {|x| (x[:classA_name] == classB or @subset_table[x[:classA_name]] == classB) and
					(x[:classB_name] == classA or @subset_table[x[:classB_name]] == classA or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classA)) ) }
			end
			
			if not relation
				puts "Error: Could not find relations between specified classes " +
					"(#{classA} and #{classB})."
				return
			end
			
			
			isM2Many = (relation[:relation_type] == RelationType::MANY_TO_MANY)
			polyLine = isPolyB ? "(is#{classB} b)" : nil
			notPostB = "(not (Post_#{classB} b))"
			classB = "PolymorphicClass" if isPolyB
			related = nil
			
			if not isM2Many and not switched 
				related = "(= b (#{relation[:relation_name]} x))"
			elsif not isM2Many and switched 
				related = "(= x (#{relation[:relation_name]} b))"
			elsif isM2Many and not switched 
				related = "(#{relation[:relation_name]} x b)"
			else
				related = "(#{relation[:relation_name]} b x)"
			end
			
			
			@property = "(forall ((b #{classB}))"
			@property << "\n  (=> #{polyLine}" if polyLine
			@property << "\n    (=> #{related}  #{notPostB} )"
			@property << "\n  )" if polyLine
			@property << "\n)"	
			
			proj_relations = [relation[:relation_name]]
			
		
		elsif property_info[0] == "deletePropagates2"
			
			classA = property_info[1]
			classB = property_info[2]
			classC = property_info[3]
			proj_classes = [classA, classB, classC]
			isPolyB = @polymorphic_parentsmap.has_key? classB
			isPolyC = @polymorphic_parentsmap.has_key? classC
			switched1 = false
			switched2 = false
			
			relation1 = @relations.find {|x| (x[:classA_name] == classA or @subset_table[x[:classA_name]] == classA) and
					(x[:classB_name] == classB or @subset_table[x[:classB_name]] == classB or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classB)) ) }
			if relation1.nil?
				switched1 = true
				relation1 = @relations.find {|x| (x[:classA_name] == classB or @subset_table[x[:classA_name]] == classB) and
					(x[:classB_name] == classA or @subset_table[x[:classB_name]] == classA or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classA)) ) }
			end
			relation2 = @relations.find {|x| (x[:classA_name] == classB or @subset_table[x[:classA_name]] == classB) and
					(x[:classB_name] == classC or @subset_table[x[:classB_name]] == classC or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classC)) ) }
			if relation2.nil?
				switched2 = true
				relation2 = @relations.find {|x| (x[:classA_name] == classC or @subset_table[x[:classA_name]] == classC) and
					(x[:classB_name] == classB or @subset_table[x[:classB_name]] == classB or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classB)) ) }
			end
			
			if not relation1 or not relation2
				puts "Error: Could not find relations between specified classes " +
					"(#{classA} and #{classB}, #{classB} and #{classC})."
				return
			end
			
			polyLine = nil
			if isPolyB and isPolyC
				polyLine = "(and (is#{classB} b) (is#{classC} c))"
			elsif isPolyB
				polyLine = "(is#{classB} b)"
			elsif isPolyC
				polyLine = "(is#{classC} c)"
			end
			
			isM2Many1 = (relation1[:relation_type] == RelationType::MANY_TO_MANY)
			isM2Many2 = (relation2[:relation_type] == RelationType::MANY_TO_MANY)
			notPostC = "(not (Post_#{classC} c))"
			classB = "PolymorphicClass" if isPolyB
			classC = "PolymorphicClass" if isPolyC
			related1 = nil
			related2 = nil
			
			if not isM2Many1 and not switched1 
				related1 = "(= b (#{relation1[:relation_name]} x))"
			elsif not isM2Many1 and switched1 
				related1 = "(= x (#{relation1[:relation_name]} b))"
			elsif isM2Many1 and not switched1 
				related1 = "(#{relation1[:relation_name]} x b)"
			else
				related1 = "(#{relation1[:relation_name]} b x)"
			end
			if not isM2Many2 and not switched2 
				related2 = "(= c (#{relation2[:relation_name]} b))"
			elsif not isM2Many2 and switched2 
				related2 = "(= b (#{relation2[:relation_name]} c))"
			elsif isM2Many2 and not switched2 
				related2 = "(#{relation2[:relation_name]} b c)"
			else
				related2 = "(#{relation2[:relation_name]} c b)"
			end
			

			@property = "(forall ((b #{classB})(c #{classC}))"
			@property << "\n  (=> #{polyLine}" if polyLine
			@property << "\n    (=> (and #{related1} #{related2}) #{notPostC} )"
			@property << "\n  )" if polyLine
			@property << "\n)"	
			
			proj_relations = [relation1[:relation_name], relation2[:relation_name]]


		elsif property_info[0] == "noDeletePropagation"
			
			classA = property_info[1]
			classB = property_info[2]
			proj_classes = [classA, classB]
			isPolyB = @polymorphic_parentsmap.has_key? classB
			switched = false
			
			relation = @relations.find {|x| (x[:classA_name] == classA or @subset_table[x[:classA_name]] == classA) and
					(x[:classB_name] == classB or @subset_table[x[:classB_name]] == classB or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classB)) ) }
			if relation.nil?
				switched = true
				relation = @relations.find {|x| (x[:classA_name] == classB or @subset_table[x[:classA_name]] == classB) and
					(x[:classB_name] == classA or @subset_table[x[:classB_name]] == classA or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classA)) ) }
			end
			
			if not relation
				puts "Error: Could not find relations between specified classes " +
					"(#{classA} and #{classB})."
				return
			end
			
			
			isM2Many = (relation[:relation_type] == RelationType::MANY_TO_MANY)
			polyLine = isPolyB ? "(is#{classB} b)" : nil
			postB = "(Post_#{classB} b)"
			classB = "PolymorphicClass" if isPolyB
			related = nil
			
			if not isM2Many and not switched 
				related = "(= b (#{relation[:relation_name]} x))"
			elsif not isM2Many and switched 
				related = "(= x (#{relation[:relation_name]} b))"
			elsif isM2Many and not switched 
				related = "(#{relation[:relation_name]} x b)"
			else
				related = "(#{relation[:relation_name]} b x)"
			end
			
			
			@property = "(forall ((b #{classB}))"
			@property << "\n  (=> #{polyLine}" if polyLine
			@property << "\n    (=> #{related}  #{postB} )"
			@property << "\n  )" if polyLine
			@property << "\n)"	
			
			proj_relations = [relation[:relation_name]]
			
		
		elsif property_info[0] == "noDeletePropagation2"
			
			classA = property_info[1]
			classB = property_info[2]
			classC = property_info[3]
			proj_classes = [classA, classB, classC]
			isPolyB = @polymorphic_parentsmap.has_key? classB
			isPolyC = @polymorphic_parentsmap.has_key? classC
			switched1 = false
			switched2 = false
			
			relation1 = @relations.find {|x| (x[:classA_name] == classA or @subset_table[x[:classA_name]] == classA) and
					(x[:classB_name] == classB or @subset_table[x[:classB_name]] == classB or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classB)) ) }
			if relation1.nil?
				switched1 = true
				relation1 = @relations.find {|x| (x[:classA_name] == classB or @subset_table[x[:classA_name]] == classB) and
					(x[:classB_name] == classA or @subset_table[x[:classB_name]] == classA or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classA)) ) }
			end
			relation2 = @relations.find {|x| (x[:classA_name] == classB or @subset_table[x[:classA_name]] == classB) and
					(x[:classB_name] == classC or @subset_table[x[:classB_name]] == classC or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classC)) ) }
			if relation2.nil?
				switched2 = true
				relation2 = @relations.find {|x| (x[:classA_name] == classC or @subset_table[x[:classA_name]] == classC) and
					(x[:classB_name] == classB or @subset_table[x[:classB_name]] == classB or 
						(@polymorphic_classmap[x[:classB_name]] and @polymorphic_classmap[x[:classB_name]].include?(classB)) ) }
			end
			
			if not relation1 or not relation2
				puts "Error: Could not find relations between specified classes " +
					"(#{classA} and #{classB}, #{classB} and #{classC})."
				return
			end
			
			polyLine = nil
			if isPolyB and isPolyC
				polyLine = "(and (is#{classB} b) (is#{classC} c))"
			elsif isPolyB
				polyLine = "(is#{classB} b)"
			elsif isPolyC
				polyLine = "(is#{classC} c)"
			end
			
			isM2Many1 = (relation1[:relation_type] == RelationType::MANY_TO_MANY)
			isM2Many2 = (relation2[:relation_type] == RelationType::MANY_TO_MANY)
			postC = "(Post_#{classC} c)"
			classB = "PolymorphicClass" if isPolyB
			classC = "PolymorphicClass" if isPolyC
			related1 = nil
			related2 = nil
			
			if not isM2Many1 and not switched1 
				related1 = "(= b (#{relation1[:relation_name]} x))"
			elsif not isM2Many1 and switched1 
				related1 = "(= x (#{relation1[:relation_name]} b))"
			elsif isM2Many1 and not switched1 
				related1 = "(#{relation1[:relation_name]} x b)"
			else
				related1 = "(#{relation1[:relation_name]} b x)"
			end
			if not isM2Many2 and not switched2 
				related2 = "(= c (#{relation2[:relation_name]} b))"
			elsif not isM2Many2 and switched2 
				related2 = "(= b (#{relation2[:relation_name]} c))"
			elsif isM2Many2 and not switched2 
				related2 = "(#{relation2[:relation_name]} b c)"
			else
				related2 = "(#{relation2[:relation_name]} c b)"
			end
			

			@property = "(forall ((b #{classB})(c #{classC}))"
			@property << "\n  (=> #{polyLine}" if polyLine
			@property << "\n    (=> (and #{related1} #{related2}) #{postC} )"
			@property << "\n  )" if polyLine
			@property << "\n)"	
			
			proj_relations = [relation1[:relation_name], relation2[:relation_name]]

		end


		if projection_on
			return [proj_classes, proj_relations]
		else
			return nil
		end

	end
	
	
	
	# Infers the deletePropagates property about translated data model and outputs 
	#  SMT specifications for each inferred property into the current working directory
	# Performs projection if 'projection_on' is true.
	def infer_deletePropagates( projection_on )		
		# Property 2: Inferring deletePropagates[classA, classB] properties: When an
		#  object of classA is deleted, directyl related classB objects are also
		#  deleted.
		# Heuristic 2: Check property for each one-to-many or one-to-zeroOne 
		#  relation.
		translate() if not @translated
		
		num_prop2 = 1  # number of properties inferred, to name files		
		relation_graph = Graph.new true  # graph whose nodes are classes and edges are relations
					# Does NOT include conditional, transitive or many-to-many relations
					# Directed edges: From classA to classB if one-to-many rel
					#  between them (or one-to-zero/one)		
		
		# Populate 'relation_graph'
		@relations.each do |r|
			name = r[:relation_name]
			classA = r[:classA_name]
			classB = r[:classB_name]
			is_many2many = (r[:relation_type] == RelationType::MANY_TO_MANY)
			is_cond_rel = (@conditional_classes.include? classA) or (@conditional_classes.include? classB)
			is_thru_rel = @through_relations.find {|x| x[:relation_name] == name}

			next if is_many2many or is_thru_rel or is_cond_rel
					# try taking out poly.  try including cond.
					#   (but make sure the backend class levels calculation 
					#    happening correctly with these two types of rels)
					
			relation_graph.add_edge(classB, classA, r)
		
		end

		# Apply levels-checking heuristic
		class_levels = relation_graph.assign_levels

		
		# Infer properties
		@relations.each do |rel|		
			classA = rel[:classA_name]
			classB = rel[:classB_name]
			relname = rel[:relation_name]
			smtclassA = classA
						
			next if ( rel[:relation_type] == RelationType::MANY_TO_MANY)
			next if (@conditional_classes.include? classA) or (@conditional_classes.include? classB)
			next if @through_relations.find {|x| x[:relation_name] == relname}
			next if (class_levels[classA] - class_levels[classB]).abs != 1
			
			# Get smt class for classB
			if @polymorphic_classmap.has_key? classB
				# classB is a poly superclass.  choose a subclass to represent group
				classB = @polymorphic_classmap[classB][0]
			end
			
			# Get 'smtclassA'
			if not @top_level_sorts.include? classA
				if @polymorphic_parentsmap.has_key? classA
					smtclassA = "PolymorphicClass"
				else
					smtclassA = @subset_table[classA]
				end
			end
			
			# Create property:
			#   [deleteB] => related A's are also deleted, i.e. related A's not in Post
			@property = ";Property: deletePropagation[#{classB}, #{classA}, #{relname}]\n"
			@property << ";Repair: Add the ':dependent => :destroy' option to the association in the #{classB} class.\n"
			@property << "(forall ((y #{smtclassA}))"
			@property << "\n  (=> "
			@property << "(and (is#{classA} y) "  if smtclassA != classA
			@property << "(= x (#{relname} y))"
			@property << " )"  if smtclassA != classA
			@property << "\n      (not (Post_#{classA} y))"
			@property << "\n) )"	

			# Print SMT spec
			proj_classes = [classA, classB]
			proj_relations = [relname]
			filename = "deletePropagation_" + num_prop2.to_s + ".smt"
			num_prop2 += 1	
			
			if projection_on
				print_smt_spec(filename, classB, proj_classes, proj_relations)
			else
				print_smt_spec(filename, classB)
			end
			@property = nil
		end
		
	end
	
	
	
	# Infers the noOrphans property about translated data model and outputs 
	#  SMT specifications for each inferred property into the current working directory
	# Performs projection if 'projection_on' is true.
	def infer_noOrphans( projection_on )
		# Property 5: Inferring noOrphans[classA, classB] properties: When classA
		#  is a class with only one relation which is to classB, and an object of 
		#  classB is deleted, related objects of classA are not orphaned, i.e. they
		#  are deleted too.
		# Heuristic 5: Check for all classes with only one outgoing relation
		translate() if not @translated
		
		num_prop5 = 1  		# number of properties inferred
		relations_byclass = {}  # hash that maps a class to all of its relations
		subclass_table = {}     # hash that maps a class to all its children classes
		subclasses = @subset_table.keys - @conditional_classes  # list of inheritance classes
											# (these are NOT in top_level_sorts)
		polyclasses = @polymorphic_classmap.keys
		

		# Initialize 'relations_byclass' for all Rails classes
		@top_level_sorts.each do |c|  # non-polymorphic classes with ActiveRecord as base class
			relations_byclass[c] = []
		end
		subclasses.each do |c|	   # non-polymorphic classes with non-ActiveRecord base classes
			relations_byclass[c] = []
		end	
		@polymorphic_parentsmap.each do |c, polyparents|
			relations_byclass[c] = []  # classes with polymorphic relations
		end	

		# Populate 'subclass_table' hash
		subclasses.each do |c|
			p = @subset_table[c]
			if subclass_table.has_key? p
				subclass_table[p].push c
			else
				subclass_table[p] = [ c ]
			end			
		end

		# Populate 'relations_byclass'
		@relations.each do |r|
			name = r[:relation_name]
			classA = r[:classA_name]
			classB = r[:classB_name]
			classA = @subset_table[classA] if @conditional_classes.include? classA
			classB = @subset_table[classB] if @conditional_classes.include? classB
			#cond_rel = (@conditional_classes.include? classA) or (@conditional_classes.include? classB)
			is_thru_rel = @through_relations.find {|x| x[:relation_name] == name}
			is_poly_rel = polyclasses.include? classB
			
	
			if is_poly_rel  # copy relation in each poly-subclass
				@polymorphic_classmap[classB].each do |c|
					relations_byclass[c] << r
				end
			else
				relations_byclass[classB] << r
			end 
			
			
			if subclass_table.has_key? classB
				subclass_table[classB].each do |c| #copyrel in any children classes
					relations_byclass[c] << r
				end
			end
			
			next if is_thru_rel  # :through relations are one-way, so only store in classB
			
			relations_byclass[classA] << r
			
			if subclass_table.has_key? classA
				subclass_table[classA].each do |c| #copy rel in any children classes
					relations_byclass[c] << r
				end
			end
			
			
			# Note: the following copies pointers to relation objects (doesn't copy object itself)
			## relations_byclass[classA] << r
		end		
		
		
		# Infer properties
		relations_byclass.each_key do |c|
			if relations_byclass[c].length == 1
				# when delete objects of related class, this class also deleted
				#  i.e. [deleteB] => related A's are deleted
				
				rel = relations_byclass[c][0]
				relname = rel[:relation_name]
				classA = rel[:classA_name]
				classB = rel[:classB_name]
				subclass_assert = nil
				is_classA = nil  # true if 'c' is classA side of relation
				smtclass = nil
				
				next if rel[:relation_type] == RelationType::MANY_TO_MANY
				
				# Populate 'is_classA', 'smtclass', and 'subclass_assert'
				if c == classA 
					is_classA = true
					smtclass = c
				elsif c == classB
					next  # want B has_many A's
				elsif @subset_table[c] == classA
					is_classA = true
					smtclass = classA
					subclass_assert = "(is#{c} y)"
				elsif @subset_table[c] == classB  # classB parent of c
					next # want B has_many A's
				elsif c == @subset_table[classA] 
					is_classA = true
					smtclass = c
					subclass_assert = "(is#{classA} y)"
				elsif c == @subset_table[classB]  # classB conditional subclass of c
					next # want B has_many A's
				elsif @polymorphic_classmap.has_key?(classA) and @polymorphic_classmap[classA].include?(c) 
					is_classA = true
					smtclass = "PolymorphicClass"
					subclass_assert = "(is#{c} y)"
				else # classB polymorphic superclass of c
					next # want B has_many A's
				end
				
				# Populate 'del_obj_type'
				del_obj_type = [classB]
				if not @top_level_sorts.include? classB and not @polymorphic_parentsmap.has_key? classB
					if @subset_table[classB]  # classB conditional subclass of another class
						del_obj_type = [ @subset_table[classB] ]
					elsif @polymorphic_classmap.has_key?(classB)  # class is a polymorphic superclass
						del_obj_type = @polymorphic_classmap[classB]
						# in this case, we want to create a property for every polymorphic subclass
					else #error!
						puts "error in infer_noOrphans() " 
							#+ "(Error info: #{del}, #{classA}, #{classB}"
						del_obj_type = []
					end
				end
				
				# Write property
				related = "(= x (#{relname} y))"
				property = "  (forall ((y #{smtclass}))"
				property << "\n    (=> "
				property << "\n      (and #{subclass_assert}" if subclass_assert
				property << "   #{related}"
				property << "  )" if subclass_assert
				property << "\n      (not (Post_#{c} y))"
				property << "\n  ) )"	
			
				# Print SMT spec with property
				proj_relations = [relname]
				del_obj_type.each do |del_obj|
					@property = ";Property: orphanPrevention[#{del_obj}, #{c}]\n"
					@property << ";Repair: Add the ':dependent => :destroy' option to the association in the #{del_obj} class.\n"
					@property << property 
					proj_classes = [c, del_obj]
					filename = "orphanPrevention_" + num_prop5.to_s + ".smt"
					num_prop5 += 1
					
					if projection_on
						print_smt_spec(filename, del_obj, proj_classes, proj_relations)
					else
						print_smt_spec(filename, del_obj)
					end
				end
				@property = nil
			end
		end

	end
	
	
	
	# Infers the transitive property about translated data model and outputs 
	#  SMT specifications for each inferred property into the current working directory
	# Performs projection if 'projection_on' is true.
	def infer_transitive( projection_on )
		# Property 4: Inferring transitive[classA, classB, classC] properties: Checks
		#  that the relations between the given classes have the transitivity
		#  property (navigating across two is the same as navigating one).
		# Heuristic 4: Infer for all class triples with a triangle relationship.
		#  Specifically, classA is the one with two has_one/has_many
		#  declarations and classC is the one with two belongs_to declarations.
		#  (:through option can only be set on a has_many or has_one
		#  declaration)
		translate() if not @translated
		
		num_prop4 = 1	# number of properties inferred
		polyclasses = @polymorphic_classmap.keys
		relation_graph = Graph.new true  # graph whose nodes are classes and edges are relations
					# Does NOT include conditional, polymorphic or many-to-many relations
					#  nor transitive relations
					# Directed edges: From classA to classB if one-to-many rel
					#  between them (or one-to-zero/one)
		pairs_added = []	# to store whether a relation between two classes has already been 
					#  added to 'relation_graph'
		
		# Populate 'relation_graph'
		@relations.each do |r|
			name = r[:relation_name]
			classA = r[:classA_name]
			classB = r[:classB_name]
			is_many2many = (r[:relation_type] == RelationType::MANY_TO_MANY)
			is_cond_rel = (@conditional_classes.include? classA) or (@conditional_classes.include? classB)
			is_thru_rel = @through_relations.find {|x| x[:relation_name] == name}
			is_poly_rel = polyclasses.include? classB
		
			next if is_many2many or is_thru_rel or is_cond_rel or is_poly_rel
			
			if not pairs_added.include? [classB, classA]  # only add one relation per class pair
				pairs_added << [classB, classA]
				relation_graph.add_edge(classB, classA, r)
			end
		end
		
		
		# Infer properties -TYPE 2 TRANSITIVE TRIANGLES
		triangles_found = []  # stores all type2 triangles found, so as to not repeat when 
						# searching for type1 triangles
		relation_graph.find_triangles2.each do |c1, c2, c3, r12, r23, r13|	
			# Setup
			proj_classes = [c1, c2, c3] 
			proj_relations = [r12[:relation_name], r23[:relation_name], 
			  r13[:relation_name]]
			subclass_stmts = ""  # stmts to enforce domains, if needed.
							# e.g. "(isPolySubclass y) (is..."
			smtclass = {}  # stores the classes c1, c2 and c3 as 
						# encoded in SMT, e.g. "PolymorphicClass"
							
			for i in 1..3
				c = proj_classes[i-1]
				if @polymorphic_parentsmap.has_key? c
					smtclass[c] = "PolymorphicClass"
					subclass_stmts << " (is#{c} y#{i})"
				elsif not @top_level_sorts.include? c
					smtclass[c] = @subset_table[c]
					subclass_stmts << " (is#{c} y#{i})"
				else
					smtclass[c] = c
				end
			end

			related1 = "(= y1 (#{r12[:relation_name]} y2))"
			related2 = "(= y2 (#{r23[:relation_name]} y3))"
			related3 = "(= y1 (#{r13[:relation_name]} y3))"


			# Write property for type 2 triangles
			@property = ";Property: transitive2[#{c1}, #{c2}, #{c3}, #{r12[:relation_name]}, #{r23[:relation_name]}, #{r13[:relation_name]}]\n"
			@property << ";Repair: Add the ':through => :#{c2.underscore.pluralize}' option to the association in the #{c1} class. "
			@property << "Remove the 'belongs_to :#{c1.underscore}' in the #{c3} class as well.\n"
			@property << "(assert (not (forall "
			@property << "((y1 #{smtclass[c1]})(y2 #{smtclass[c2]})(y3 #{smtclass[c3]}))"
			@property << "\n  (=> (and "
			@property << "\n       #{subclass_stmts}"  if subclass_stmts.length > 0
			@property << "\n       #{related1} #{related2} )"
			@property << "\n   #{related3})"
			@property << "\n)))"			
			
			
			# Create SMT spec
			filename = "transitive_" + num_prop4.to_s + ".smt"
			num_prop4 += 1
			if projection_on
				print_smt_spec(filename, nil, proj_classes, proj_relations)
			else
				print_smt_spec(filename)
			end
			@property = nil
			
			# store triangles found so no repeats added during triangle1 search
			triangles_found << [c1, c2, c3, r12[:relation_name], r23[:relation_name], r13[:relation_name]].sort  
		end
		
		
		# Infer properties -TYPE 1 TRANSITIVE TRIANGLES
		relation_graph.find_triangles1.each do |c1, c2, c3, r12, r32, r13|
			next if triangles_found.include? [c1, c2, c3, r12[:relation_name], r32[:relation_name], r13[:relation_name]].sort 
			
			# Setup
			proj_classes = [c1, c2, c3] 
			proj_relations = [r12[:relation_name], r32[:relation_name], 
				r13[:relation_name]]
			subclass_stmts = ""  # stmts to enforce domains, if needed.
							# e.g. "(isPolySubclass y) (is..."
			smtclass = {}  # stores the classes c1, c2 and c3 as 
						# encoded in SMT, e.g. "PolymorphicClass"
							
			for i in 1..3
				c = proj_classes[i-1]
				if @polymorphic_parentsmap.has_key? c
					smtclass[c] = "PolymorphicClass"
					subclass_stmts << " (is#{c} y#{i})"
				elsif not @top_level_sorts.include? c
					smtclass[c] = @subset_table[c]
					subclass_stmts << " (is#{c} y#{i})"
				else
					smtclass[c] = c
				end
			end

			related1 = "(= y1 (#{r12[:relation_name]} y2))"
			related2 = "(= y3 (#{r32[:relation_name]} y2))"
			related3 = "(= y1 (#{r13[:relation_name]} y3))"


			# Write property for type 1 triangles
			@property = ";Property: transitive1[#{c1}, #{c2}, #{c3}]\n"
			@property << ";Repair: Add the ':through => :#{c2.underscore.pluralize}' option to the association in the #{c1} class. "
			@property << "Remove the 'belongs_to :#{c1.underscore}' in the #{c3} class as well.\n"
			@property << "(assert (not (forall "
			@property << "((y1 #{smtclass[c1]})(y2 #{smtclass[c2]})(y3 #{smtclass[c3]}))"
			@property << "\n  (=> (and "
			@property << "\n       #{subclass_stmts}"  if subclass_stmts.length > 0
			@property << "\n       #{related1} #{related2} )"
			@property << "\n   #{related3})"
			@property << "\n)))"			
			
			
			# Create SMT spec
			filename = "transitive_" + num_prop4.to_s + ".smt"
			num_prop4 += 1
			if projection_on
				print_smt_spec(filename, nil, proj_classes, proj_relations)
			else
				print_smt_spec(filename)
			end
			@property = nil			
		end
		
		
		# Infer properties - TRANSITIVE N-GONS (N > 3)
		ngons = relation_graph.find_transitive_ngons
		  #ngons{[edge0_n, node0]} => [[edge0_1, node1], [edge1_2, node2], [edge2_3, node3], ]
		
		ngons.each do |trans_info, path|
			# Setup
			trans_rel = trans_info[0][:relation_name]
			curr_node = trans_info[1]			
			proj_classes = [ curr_node ]
			proj_relations = [ trans_rel ]
			subclass_stmts = ""  	# stmts to enforce domains, if needed.
							# e.g. "(isPolySubclass y) (is..."
			i = 0 				# indices for creating class variables
			class_vars = ""		# series of "(t1 SMTclass)" etc
			related = ""		# series of "(= t0 (relname1 t1))" etc
			first_class = curr_node	# stores the first,
			last_class = nil		#  last, and 
			sec_to_last_class = nil	#  second-to-last classes in the n-gon


			if @polymorphic_parentsmap.has_key? curr_node
				class_vars << "(t0 " << "PolymorphicClass)"
				subclass_stmts << " (is#{curr_node} t0)"
			elsif not @top_level_sorts.include? curr_node
				class_vars << "(t0 " << @subset_table[curr_node] << ")"
				subclass_stmts << " (is#{curr_node} t0)"
			else
				class_vars << "(t0 " << curr_node << ")"
			end
			
			prop_classes = curr_node + ", "
			prop_rels = ""

			path.each do |rel, next_node|
				i += 1
				proj_classes << next_node
				proj_relations << rel[:relation_name]

				prop_classes << next_node << ", "
				prop_rels << rel[:relation_name] << ", "
			
				sec_to_last_class = last_class
				last_class = next_node
				
				if @polymorphic_parentsmap.has_key? next_node
					class_vars << "(t#{i} " << "PolymorphicClass)"
					subclass_stmts << " (is#{next_node} t#{i})"
				elsif not @top_level_sorts.include? next_node
					class_vars << "(t#{i} " << @subset_table[next_node] << ")"
					subclass_stmts << " (is#{next_node} t#{i})"
				else
					class_vars << "(t#{i} " << next_node << ")"
				end
				
				related << "(= t#{i-1} (#{rel[:relation_name]} t#{i})) "
				
				curr_node = next_node
			end

			prop_rels << trans_rel
			related_trans = "(= t0 (#{trans_rel} t#{i}))"


			# Write property for transitive ngons
			@property = ";Property: transitiveN[#{prop_classes}#{prop_rels}]\n"
			@property << ";Repair: Add the ':through => :#{sec_to_last_class.underscore.pluralize}' option to the association in the #{first_class} class. "
			@property << "Create the ':#{sec_to_last_class.underscore.pluralize}' association (with the :through option set), if required. "
			@property << "Remove the 'belongs_to :#{first_class.underscore}' in the #{last_class} class as well.\n"
			@property << "(assert (not (forall \n"
			@property << "  (#{class_vars})\n"			
			@property << "(=> (and "
			@property << "\n  #{subclass_stmts}"  unless subclass_stmts.empty?
			@property << "\n  #{related} )"
			@property << "\n  #{related_trans})"
			@property << "\n)))"			
			
			
			# Create SMT spec
			filename = "transitive_" + num_prop4.to_s + ".smt"
			num_prop4 += 1
			if projection_on
				print_smt_spec(filename, nil, proj_classes, proj_relations)
			else
				print_smt_spec(filename)
			end
			@property = nil
		end

	end
	
	
	
	# Infers properties about translated data models and outputs SMT specifications for each 
	#  property in the 'output_directory'
	# Performs projection if 'projection_on' is true
	def property_inference( output_directory="inferred_properties", projection_on=false )
		start_directory = Dir.getwd 

		# Create directory to put SMT specs in (containing inferred properties)		
		if File.directory? output_directory 
			#remove all containing files
			Dir.chdir output_directory
			Dir.foreach(".") {|f| File.delete(f) if !File.directory? f }
		else 	# create it
			Dir.mkdir output_directory
			Dir.chdir output_directory
		end			
		
		# Infer properties
		infer_transitive( projection_on )
		infer_noOrphans( projection_on )
		infer_deletePropagates( projection_on )

		Dir.chdir start_directory  # change back to original directory		
	end
	
	# Creates an entity-relation diagram and outputs it to 'filename'
	# Output as a .dot file, unless 'png' set to true
	def create_erd( filename, png=false )
		translate() if not @translated

		begin
			# Initialize structures needed to create ERD (aka schema graph)
			schema = GraphViz.new( :myschema, :type => :graph ) 
			nodes = {}  		# nodes in schema graph
			polyclasses = @polymorphic_classmap.keys
			subclasses = @subset_table.keys - @conditional_classes  
				# list of inheritance classes (these are NOT in top_level_sorts)

			# Initialize 'schema' graph
			schema[:rankdir] = :RL
			schema.node[:shape] = :box
			schema.node[:fontsize] = 11
			schema.node[:fontname] = "Arial"
			schema.edge[:arrowhead] = :none
			schema.edge[:arrowtail] = :normal
			schema.edge[:dir] = :both
			schema.edge[:fontsize] = 9
			schema.edge[:fontname] = "Arial"

			# Populate nodes of 'schema' graph
			@top_level_sorts.each do |c|  # non-polymorphic classes with ActiveRecord as base class
				nodes[c] = schema.add_node c
			end
			subclasses.each do |c|	   # non-polymorphic classes with non-ActiveRecord base classes
				nodes[c] = schema.add_node c
			end	
			polyclasses.each do |pc|
				nodes[pc] = schema.add_node( pc, {:shape => :diamond} )
			end
			
			# Add polymorphic relation edges to 'schema' graph
			@polymorphic_parentsmap.each do |c, polyparents|
				nodes[c] = schema.add_node c
				polyparents.each do |p|
					schema.add_edge( nodes[p], nodes[c], {:arrowtail => :none, :color => :grey60} ) 
				end
			end	
			
			# Add inheritance edges to 'schema' graph
			subclasses.each do |c|		
				schema.add_edge( nodes[ @subset_table[c] ], nodes[c], {:arrowtail => :odiamond} ) 
			end

			# Add remaining 'schema' graph edges
			@relations.each do |r|
				name = r[:relation_name]
				classA = r[:classA_name]
				classB = r[:classB_name]
				cond_rel = (@conditional_classes.include? classA) or (@conditional_classes.include? classB)
				cond_relA = classA if @conditional_classes.include? classA
				cond_relB = classB if @conditional_classes.include? classB
				classA = @subset_table[classA] if @conditional_classes.include? classA
				classB = @subset_table[classB] if @conditional_classes.include? classB
				is_thru_rel = @through_relations.find {|x| x[:relation_name] == name}
				edge = nil
				

				if r[:relation_type] == RelationType::MANY_TO_MANY
					edge = schema.add_edge( nodes[classA], nodes[classB], {:arrowhead => :normal} ) 
				elsif r[:relation_type] == RelationType::ONE_TO_ONE
					edge = schema.add_edge( nodes[classA], nodes[classB], {:arrowtail => :dot} )
				else
					edge = schema.add_edge( nodes[classA], nodes[classB] )
				end			
				edge.style = :dashed  if is_thru_rel
				edge.label = cond_relA  if cond_relA
				edge.label = cond_relB  if cond_relB
				edge.color = :grey60  if cond_rel
				edge.fontcolor = :grey60  if cond_rel
			end
			
			# Output schema graph
			if png
				schema.output( :png => filename )		
			else
				schema.output( :dot => filename )		
			end

		rescue
			$stderr.puts "Error creating ERD."   #TODO: output exception detailso to log
			return
		end
		
		puts "ERD created and output to #{filename}"
	end


	
	# Prints the SMT-LIBv2 specification based on the private variables populated in translate()
	#   Specification is print to 'filename' if given, otherwise to Standard Output
	# If an assert stub should be printed about deleting an object, the type of the object
	#  should be passed in as a string in 'deleteobjtype'
	# If projection is to be performed, array or classes and relations to project should
	#  be passed into 'projected_classes' and 'projected_relations'
	# Also prints property if @property variable populated
	def print_smt_spec(filename, deleteobjtype = nil, projected_classes = nil, 
	  projected_relations = nil) 
		out = filename ? File.new(filename, "w") : $stdout #define output stream
		proj_relations = nil	# complete set of classes for projection
		proj_classes = nil	# complete set of relations
		delete_assert = nil	# the delete assert, if deletion to be modeled
		
		if projected_relations or projected_classes	
			proj_classes, proj_relations =  complete_projection_set( projected_classes, 
			  projected_relations, deleteobjtype)
		end 
		
		if deleteobjtype
			proj_classes, proj_relations, delete_assert = generatestub_delete_assert(
			  deleteobjtype, proj_classes, proj_relations)
		end


		out.print "\n;***** SMT2*SPECIFICATION *****\n"
		out.print "(set-logic AUFLIA)\n"
		#out.print "(set-option :model-compact true)\n"
		
		subclasses = @subset_table.keys - @conditional_classes
		polyclassmap_values = @polymorphic_parentsmap.keys
		projected_pclasses = []

		# Declare a sort for each class
		@top_level_sorts.each do |klass|
			next if proj_classes and not proj_classes.include? klass
			out.print "(declare-sort ", klass, " 0)\n"
		end
		

		
		# Print uninterpreted fxns and asserts to express polymorphic classes, if any
		polyprojected = polyclassmap_values & proj_classes if proj_classes
		if @polymorphic_classmap.length > 0 and (not proj_classes or polyprojected.length > 0)			
			out.print "(declare-sort PolymorphicClass 0)\n"
			
			@polymorphic_classmap.each_key do |pclass|				
				if proj_classes
					@polymorphic_classmap[pclass].each do |childklass|
						if proj_classes.include? childklass
							projected_pclasses.push pclass
							break
						end
					end
					
					next if not projected_pclasses.include? pclass
				end
				
				out.print "(declare-fun is", pclass, "(PolymorphicClass) Bool)\n"
			end
			
			polyclassmap_values.each do |childklass|
				next if proj_classes and not proj_classes.include? childklass
				out.print "(declare-fun is", childklass, " (PolymorphicClass) Bool)\n"
			end
	
			
			out.print "(assert (forall ((p PolymorphicClass)) (and\n"
			
			out.print "\t"
			out.print "(or "  if not proj_classes or polyprojected.length > 1
			# assert each PolymorphicClass is an instantiation of an actual class
			polyclassmap_values.each do |childklass|
				next if proj_classes and not proj_classes.include? childklass
				out.print "(is", childklass, " p) "
			end
			out.print ")"   if not proj_classes or polyprojected.length > 1
			out.print "\n"
			
			@polymorphic_classmap.each_key do |pclass|   #superclass sets are abstract, and
				next if proj_classes and not projected_pclasses.include? pclass
				subclasses = ""
				count = 0
				
				out.print "\t(=> (is", pclass, " p)  "
				@polymorphic_classmap[pclass].each do |klass|
					next if proj_classes and not proj_classes.include? klass
					subclasses << "(is" + klass + " p) " 
					count += 1
				end
				subclasses = "(or " + subclasses  + ")" if count > 1
				out.print subclasses
				out.print " )\n"
			end

			@polymorphic_classmap.each_key do |pclass|   # subclasses are subsets of superclass
				next if proj_classes and not projected_pclasses.include? pclass
				subclasses = ""
				count = 0
				
				out.print "\t(=> " 
				@polymorphic_classmap[pclass].each do |klass|
					next if proj_classes and not proj_classes.include? klass
					subclasses << "(is" + klass + " p) "
					count += 1
				end
				subclasses = "(or " + subclasses  + ")" if count > 1
				out.print subclasses
				out.print "  (is", pclass, " p) )\n"
			end
			
			if not proj_classes or polyprojected.length > 1
				polyclassmap_values.each do |klass|	# each PClass object is an instantiation of only one subclass
					next if proj_classes and not proj_classes.include? klass
					subclasses = ""
					count = 0
					out.print "\t(=> (is", klass, " p)  "
					polyclassmap_values.each do |klass2|
						next if proj_classes and not proj_classes.include? klass2
						if klass2 != klass
							subclasses << "(not (is" + klass2 + " p)) "  
							count += 1
						end
					end
					subclasses = "(and " + subclasses + ")" if count > 1
					out.print subclasses
					out.print " )\n"
				end
			end
			
			out.print ")))\n"
		end #polymorphic-asserts and fxns
		
				
		
		# Create a subset function for each subclass
		@subset_table.each_pair do |klass, parentclass|
			next if proj_classes and not proj_classes.include? parentclass
			if polyclassmap_values.include? parentclass
				out.print "(declare-fun is", klass, " (PolymorphicClass) Bool)\n"
				out.print "  (assert (forall ((p PolymorphicClass)) \n"
				out.print "    (=> (is", klass, " p) (is", parentclass, " p) ) ))\n"
			elsif @subset_table[parentclass]
				# eg have this: (declare-fun isBlog (ContentNode) Bool)
				#  wanna print: (declare-fun isBlog_2 (Blog) Bool)
				#  but actually need to print: (declare-fun isBlog_2 (ContentNode) Bool)
				parparent = @subset_table[parentclass]
				out.print "(declare-fun is", klass, " (", parparent, ") Bool)\n"
				out.print "  (assert (forall ((p ", parparent, ")) \n"
				out.print "    (and (=> (is", klass, " p) (is", parentclass, " p) )\n"
				out.print "       (=> (not (is", parentclass, " p)) (not (is", klass, " p))) )))\n"
			else
				out.print "(declare-fun is", klass, " (", parentclass, ") Bool)\n"

			end				
		end
		
		
		# Print Post_object functions
		out.print "\n"
		
		#print fxns that return true if an object exists after a deletion
		@top_level_sorts.each do |klass|
			next if proj_classes and not proj_classes.include? klass
			out.print "(declare-fun Post_", klass, " (", klass, ") Bool)\n"
		end
		# same for polymorphic classes
		if @polymorphic_classmap.length > 0 and (not proj_classes or polyprojected.length > 0)			
			polyclassmap_values.each do |childklass|
				next if proj_classes and not proj_classes.include? childklass
				out.print "(declare-fun Post_", childklass, " (PolymorphicClass) Bool)\n"
			end
		
		end
		# same for inheritance child classes
		@subset_table.each_pair do |klass, parentclass|
			next if not subclasses.include? klass
			next if proj_classes and not proj_classes.include? parentclass
			if polyclassmap_values.include? parentclass # get SMT encoding of parentclass
				parentclass = "PolymorphicClass"
			elsif @subset_table[parentclass]
				parentclass = @subset_table[parentclass]
			end
			
			out.print "(declare-fun Post_", klass, " (", parentclass, ") Bool)\n"
		end
		
		# Declare an uninterpreted fxn to express each relation
		post_relations = ""
		@relations.each do |relation|
			relname = relation[:relation_name]
			reltype = relation[:relation_type]
			realclassA = relation[:classA_name]
			realclassB = relation[:classB_name]
			classA = relation[:classA_name]
			classB = relation[:classB_name]
			subclassA = nil
			subclassB = nil
			
			next if proj_relations and not proj_relations.include? relname
			
			if not @top_level_sorts.include? realclassA
				if @polymorphic_classmap.has_key? realclassA or polyclassmap_values.include? realclassA
					classA = "PolymorphicClass"
				else
					subclassA = classA
					classA = @subset_table[classA]
					
					if @polymorphic_classmap.has_key? classA or polyclassmap_values.include? classA
						classA = "PolymorphicClass"
					end
				end
			end

			if not @top_level_sorts.include? realclassB
				if @polymorphic_classmap.has_key? realclassB or polyclassmap_values.include? realclassB
					classB = "PolymorphicClass"
				else 
					subclassB = classB
					classB = @subset_table[classB]
					
					if @polymorphic_classmap.has_key? classB or polyclassmap_values.include? classB
						realclassB = classB
						classB = "PolymorphicClass"
					end
					
				end
			end

			# Print relation as uninterpreted fxn
			case reltype
				when RelationType::ONE_TO_MANY
					out.print "(declare-fun ", relname, " (", classA, ") ", 
						classB, ")\n"
				when RelationType::ONE_TO_ONE
					out.print "(declare-fun ", relname, " (", classA, ") ", classB, ")\n"
					out.print "(assert (forall ((x1 ", classA, ") (x2 ", classA, "))\n"
					if classA == "PolymorphicClass"
						out.print "\t(=> (and (is", realclassA, " x1)  (is", realclassA, " x2))\n" 
					end
					out.print "\t(=> (not (= x1 x2))  (not (= (", relname, " x1) (", relname, " x2) ))  )\n"
					if classA == "PolymorphicClass"
						out.print "\t) \n" 
					end
					out.print "))\n"
				when RelationType::MANY_TO_MANY
					out.print "(declare-fun ", relname, " (", classA, " ", 
						classB, ") Bool)\n"
			end
					
			# If necessary, add asserts to enforce domains/ranges of relations 
			#  (for relations with non-top-level classes)
			if subclassA and reltype == RelationType::MANY_TO_MANY
				out.print "(assert (forall ((a ", classA, ") (b ", 
					classB, "))\n\t(=> (", relname, " a b) (is", 
					subclassA, " a) )\n))\n"
			end
			
			if subclassB
				if reltype == RelationType::MANY_TO_MANY
					out.print "(assert (forall ((a ", classA, ") (b ", 
						classB, "))\n\t(=> (", relname, " a b) (is", 
						subclassB, " b) )\n))\n"
				else
					out.print "(assert (forall ((a ", classA, "))\n\t",
						"(is", subclassB, " (", relname, " a))\n))\n"
				end
			end
			if classB == "PolymorphicClass"
				if reltype != RelationType::MANY_TO_MANY  
					out.print "(assert (forall ((a ", classA, "))\n\t",
						"(is", realclassB, " (", relname, " a))\n))\n"
				end
			end
						
			# If this is a through relation, add assert to enforce transitivity
			thru_relation = @through_relations.find {|x| x[:relation_name] == relname}
			if thru_relation
				tclassA = realclassB
				tclassB = thru_relation[:thruclass_name]
				tclassC = realclassA
				
				isPolyA = @polymorphic_parentsmap.has_key? tclassA
				isPolyB = @polymorphic_parentsmap.has_key? tclassB
				isPolyC = @polymorphic_parentsmap.has_key? tclassC
				relname1 = thru_relation[:relation1][:relation_name]
				relname2 = thru_relation[:relation2][:relation_name]
				relname3 = relname
				switched = false
				
				if thru_relation[:thru_reltype] == ThroughRelType::TYPE2
					switched = true
				end
				
				polyLineB = isPolyB ? "(is#{tclassB} b)" : nil
				polyLineAC = nil
				if isPolyA and isPolyC
					polyLineAC = "(and (is#{tclassA} a) (is#{tclassC} c) )"
				elsif isPolyA
						polyLineAC = "(is#{tclassA} a)"
				elsif isPolyC
						polyLineAC = "(is#{tclassC} c)"
				end
				
				related1 = "(= a (#{relname1} b))"
				related2 = switched ? "(= c (#{relname2} b))" : 
								"(= b (#{relname2} c))"
				related3 = "(= a (#{relname3} c))"
				
				tclassA = classB if subclassB
				tclassC = classA if subclassA
				tclassA = "PolymorphicClass" if isPolyA
				tclassB = "PolymorphicClass" if isPolyB
				tclassC = "PolymorphicClass" if isPolyC
			

				out.print "(assert (forall ((a #{tclassA})(c #{tclassC}))"
				out.print "\n  (=> #{polyLineAC}" if polyLineAC
				out.print "\n    (and "
				out.print "\n      (=> #{related3} "
				out.print "\n          (exists ((b #{tclassB})) (and #{related1} #{related2} #{polyLineB})))"
				out.print "\n      (=> (exists ((b #{tclassB})) (and #{related1} #{related2} #{polyLineB}))"
				out.print "\n          #{related3})"
				out.print "\n    ) "
				out.print "\n  )" if polyLineAC
				out.print "\n))\n"	
			end
			
			# create fxn that returns true if two objects are related after a delete	
			post_relations << "(declare-fun Post_" + relname + " (" + classA + " " + classB + ") Bool)\n"
		end
		
		# Print Post_relations
		out.print "\n"
		out.print post_relations
		
		
		if deleteobjtype
			out.print delete_assert
		else
			out.print "\n", @property, "\n" if not @property.nil?
		end
		
		out.print "(check-sat)\n"
		out.print "(get-model)\n"
		out.print "(exit)\n"
		out.print "\n"
		
		out.close
	end #end-fxn  print_smt_spec()
	
	
	private   # all following methods are private
	
	# Adds necessary classes and relations to complete projection set
	# Precondition: 'projected_classes' is a list of Rails classes and 
	#   'projected_relations' is a list SMT relations. Either may be nil.
	# Postcondition: Returns a tuple [classes, relations] that holds the 
	#   complete projection set of classes and relations
	def complete_projection_set( projected_classes, projected_relations, deleteobjtype )
		proj_classes = nil
		proj_relations = nil
		
		if not projected_classes
			proj_classes = []
		else
			proj_classes = Array.new(projected_classes)
		end
		
		if not projected_relations
			proj_relations = []
		else
			proj_relations = Array.new(projected_relations)
		end
		
		proj_classes << deleteobjtype  if (not proj_classes.include? deleteobjtype)
		
		# If any relation is a transitive relation, include rel1 and rel2
		projected_relations.each do |rel|
			thru_relation = @through_relations.find {|x| x[:relation_name] == rel}
			if thru_relation
				rel1 = thru_relation[:relation1][:relation_name]
				rel2 = thru_relation[:relation2][:relation_name]
				proj_relations.push rel1 if not proj_relations.include? rel1
				proj_relations.push rel2 if not proj_relations.include? rel2
			end
		end
		
		# For every relation, make sure domain and range classes are included
		proj_relations.each do |rel|
			relation = @relations.find {|x| x[:relation_name] == rel}
			if relation
				dom = relation[:classA_name]
				ran = relation[:classB_name]
				if not proj_classes.include? dom
					if @polymorphic_classmap.has_key? dom
						@polymorphic_classmap[dom].each do |c|
							proj_classes.push c
						end
					elsif @conditional_classes.include? dom
						proj_classes.push @subset_table[dom]
					else
						proj_classes.push dom
					end
				end
				if not proj_classes.include? ran
					if @polymorphic_classmap.has_key? ran
						@polymorphic_classmap[ran].each do |c|
							proj_classes.push c
						end
					elsif @conditional_classes.include? ran
						proj_classes.push @subset_table[ran]
					else
						proj_classes.push ran
					end					
				end
			end
		end
		
		# For every class, make sure superclasses are included
		sup_classes = []
		proj_classes.each do |cl|
			if (not @top_level_sorts.include? cl)  and @subset_table.has_key? cl
				sup_classes << @subset_table[cl]
			end
=begin			
				if (@polymorphic_parentsmap.has_key? relclass or @polymorphic_classmap.has_key? cl)
					relclass = "PolymorphicClass"
				else
					sup_classes << @subset_table[cl]
					if @polymorphic_parentsmap.has_key? relclass or @polymorphic_classmap.has_key? relclass
						supclass = relclass
						relclass = "PolymorphicClass"
					end
				end
			end
=end
		end		
		proj_classes += sup_classes		
		
		return proj_classes, proj_relations
	end
	
	
	# Generates an assert stub that can be used to express a property of the
	#  system after an object (of type 'objtype') has been deleted
	# Precondition: 'objtype' is a string containing the name of the Rails class of 
	#   the object to delete
	# Postcondition: The complete set of projected classes, projected relations, and
	#  assert stub is returned as a 3-tuple
	def generatestub_delete_assert(objtype, projected_classes, projected_relations)			
		subclasses = @subset_table.keys - @conditional_classes
		if (not @top_level_sorts.include? objtype and 
		    not @polymorphic_parentsmap.has_key? objtype and
		    not subclasses.include? objtype)
			$stderr.print ";warning 6: ", objtype, " not a valid class\n"
			return
		end

		# Variables
		header = ""  		   # header of assert
		post_objects = {}  	   # defns of post_object functions
		post_relations = {}  	   # definitions of post_relation functions
		footer = ""  		   # footer of assert
		
		post_object_depO = {}	# dependencies a post_object fxn has on another post_object fxn
		post_object_depR = {}	#  " ... on a post_relation fxn
		post_relation_dep = {}		# dependencies a post_relation fxn has on some post_object fxn
		processed_classes = []		# process each class and relation only once
		processed_relations = []
		dependency_stack = []		# to determine Destroy propagation
		obj_smtclass = nil		# the SMT class encoding of 'objtype'
		
		
		# Find objtype as encoded in smt
		is_subclass = false  # obj's smtclass is a subclass (boolean fxn)
		if (@polymorphic_parentsmap.has_key? objtype or
		    @polymorphic_classmap.has_key? objtype)
			obj_smtclass = "PolymorphicClass"
			is_subclass = true
		elsif @subset_table[objtype]
			obj_smtclass = @subset_table[objtype]
			is_subclass = true
		else
			obj_smtclass = objtype
		end
		
		
		# Header of assert
		header << "(assert (not (forall ((x " << obj_smtclass << "))"
		header << " ;delete " << objtype  if is_subclass
		header << "\n(=> (and  "
		header << "\n\t(is" << objtype << " x)" if is_subclass
		header << "\n"
	
		# Footer of assert
		footer << ")\n"
		if @property.nil?
			footer << "\t;Insert property to check here\n"
		else
			footer << "\n" << @property << "\n"
		end
		footer << "))))\n"

		# Remove object to delete from Post_object()
		fxn_body = "\t(forall ((a " + obj_smtclass + "))\n" +
			"\t\t(ite (= a x) (not (Post_" + objtype + 
			" a)) (Post_" + objtype + " a)  ))\n"
		post_objects[objtype] = fxn_body
		processed_classes.push objtype
			
		#Remove associated objects if :dependent options set on any of its rels			
		@dependent_rels[objtype].each do |d|
			relclass = d[:relation_with]
			subclassB = nil
			relpoly = (@polymorphic_parentsmap.has_key? relclass or
					@polymorphic_classmap.has_key? relclass) ?
					true : false

			if @subset_table[relclass]
				subclassB = relclass 
				relclass = @subset_table[relclass]
				relpoly = (@polymorphic_parentsmap.has_key? relclass or
					@polymorphic_classmap.has_key? relclass) ?
					true : false
			end
			processed_classes.push relclass
			
			fxn_body = ""
			if not relpoly
				fxn_body << "\t(forall ((b " << relclass << "))\n"
			else
				fxn_body << "\t(forall ((b PolymorphicClass))\n"
			end
			fxn_body << "\t\t(ite "
			
			if d[:argument]
				fxn_body << "(= b (" << d[:relation_name] << " x)) "
			else
				if not subclassB
					fxn_body << "(= x (" << d[:relation_name] << " b)) "
				else
					fxn_body << "(and (is" << subclassB << " b) (= x (" <<  
						d[:relation_name] << " b)) ) "
				end
			end
			
			fxn_body << "(not (Post_" << relclass << " b)) (Post_" << relclass <<
				" b)  ))\n"
				
			post_objects[relclass] = fxn_body
			post_object_depO[relclass] = objtype
			post_object_depR[relclass] = d[:relation_name]
			
			if d[:dependency_type] == DependencyType::DESTROY
				dependency_stack.push relclass
			end
		end
		# repeat for depedent polymorphically-related objects (i.e. when objtype is a PolymorphicClass in spec)
		if @polymorphic_parentsmap.has_key? objtype
			@polymorphic_parentsmap[objtype].each do |superclass|
				@dependent_rels[superclass].each do |d|
					relclass = d[:relation_with]
					subclassB = nil
					relpoly = (@polymorphic_parentsmap.has_key? relclass or
							@polymorphic_classmap.has_key? relclass) ?
							true : false

					if @subset_table[relclass]
						subclassB = relclass 
						relclass = @subset_table[relclass]
						relpoly = (@polymorphic_parentsmap.has_key? relclass or
							@polymorphic_classmap.has_key? relclass) ?
							true : false
					end
					processed_classes.push relclass
					
					fxn_body = ""
					if not relpoly
						fxn_body << "\t(forall ((b " << relclass << "))\n"
					else
						fxn_body << "\t(forall ((b PolymorphicClass))\n"
					end
					fxn_body << "\t\t(ite "
					
					if d[:argument]
						fxn_body << "(= b (" << d[:relation_name] << " x)) "
					else
						if not subclassB
							fxn_body << "(= x (" << d[:relation_name] << " b)) "
						else
							fxn_body << "(and (is" << subclassB << " b) (= x (" << d[:relation_name] << " b)) ) "
						end
					end
					
					fxn_body << "(not (Post_" << relclass << " b)) (Post_" << relclass << " b)  ))\n"
					
					post_objects[relclass] = fxn_body
					post_object_depO[relclass] = objtype
					post_object_depR[relclass] = d[:relation_name]
					
					if d[:dependency_type] == DependencyType::DESTROY
						dependency_stack.push relclass
					end
				end
			end
		end
		
		#Propagate delete across dependent objects
		while not dependency_stack.empty?
			klass = dependency_stack.pop

			@dependent_rels[klass].each do |d|
				relclass = d[:relation_with]
				subclassC = nil
				relpoly = (@polymorphic_parentsmap.has_key? relclass or
						@polymorphic_classmap.has_key? relclass) ?
						true : false

				if @subset_table[relclass]
					relclass = @subset_table[relclass]
					relpoly = (@polymorphic_parentsmap.has_key? relclass or
						@polymorphic_classmap.has_key? relclass) ?
						true : false
				end

				next if processed_classes.include? relclass
				processed_classes.push relclass
			
				fxn_body = ""
				if not relpoly
					fxn_body << "\t(forall ((c " << relclass << "))\n"
				else
					fxn_body << "\t(forall ((c PolymorphicClass))\n"
				end
				fxn_body << "\t\t(ite "
				
				if subclassC
					fxn_body << "(or (not (is" << subclassC << " c))  "
				end
				
				if d[:argument]
					pklass = (@polymorphic_parentsmap.has_key? klass) ? "PolymorphicClass" : klass
					fxn_body << "(exists ((b " << pklass << ")) (and (= c (" << d[:relation_name] << " b)) "
					fxn_body << "(not (Post_" << klass << " b)) )) "		
				else
					fxn_body << "(not (Post_" << klass << " (" << d[:relation_name] << " c))) "
				end
				
				fxn_body << ")" if subclassC
				fxn_body << "\n\t\t    "
				fxn_body << "(not (Post_" << relclass << " c)) (Post_" << relclass << " c)  ))\n"
				post_objects[relclass] = fxn_body
				post_object_depO[relclass] = klass
				post_object_depR[relclass] = d[:relation_name]
				
				if d[:dependency_type] == DependencyType::DESTROY
					dependency_stack.push relclass
				end
			end
			# repeat for depedent polymorphically-related objects (i.e. type PolymorphicClass)
			if @polymorphic_parentsmap.has_key? klass
				@polymorphic_parentsmap[klass].each do |superclass|
					@dependent_rels[superclass].each do |d|
						relclass = d[:relation_with]
						subclassC = nil
						relpoly = (@polymorphic_parentsmap.has_key? relclass or
								@polymorphic_classmap.has_key? relclass) ?
								true : false

						if @subset_table[relclass]
							relclass = @subset_table[relclass]
							relpoly = (@polymorphic_parentsmap.has_key? relclass or
								@polymorphic_classmap.has_key? relclass) ?
								true : false
						end
						next if processed_classes.include? relclass
						processed_classes.push relclass
					
						fxn_body = ""
						if not relpoly
							fxn_body << "\t(forall ((c " << relclass << "))\n"
						else
							fxn_body << "\t(forall ((c PolymorphicClass))\n"
						end
						fxn_body << "\t\t(ite "
						
						if subclassC
							fxn_body << "(or (not (is" << subclassC << " c))  "
						end
						
						if d[:argument]
							fxn_body << "(exists ((b PolymorphicClass)) (and (= c (" << d[:relation_name] << " b)) "
							fxn_body << "(not (Post_" << klass << " b)) )) "
						else
							fxn_body << "(not (Post_" << klass << " (" << d[:relation_name] << " c))) "
						end
						
						fxn_body << ")" if subclassC
						fxn_body << "\n\t\t    "
						fxn_body << "(not (Post_" << relclass << " c)) (Post_" << relclass << " c)  ))\n"
						post_objects[relclass] = fxn_body
						post_object_depO[relclass] = klass
						post_object_depR[relclass] = d[:relation_name]
						
						if d[:dependency_type] == DependencyType::DESTROY
							dependency_stack.push relclass
						end
					end
				end
			end
		end

		# For any remaining sorts, define Post_object() to be true for all objects
		@top_level_sorts.each do |khlass|
			if not processed_classes.include? khlass
				fxn_body = "\t(forall ((a " << khlass << "))  (Post_" << khlass << " a) )\n"
				post_objects[khlass] = fxn_body
			end
		end
		@polymorphic_parentsmap.keys.each do |khlass|
			if not processed_classes.include? khlass
				fxn_body = "\t(forall ((a PolymorphicClass))  (Post_" << khlass << " a) )\n"
				post_objects[khlass] = fxn_body
			end
		end
		subclasses.each do |khlass|
			if not processed_classes.include? khlass	
				parentclass = @subset_table[khlass]

				# get SMT encoding of parentclass
				if @polymorphic_parentsmap.has_key? parentclass 
					parentclass = "PolymorphicClass"
				elsif @subset_table[parentclass]
					parentclass = @subset_table[parentclass]
				end
				
				fxn_body = "\t(forall ((a " << parentclass << "))  (Post_" << khlass << " a) )\n"
				post_objects[khlass] = fxn_body
			end
		end
		
		
		#Define Post_Relation() for belongs_to relations in classes with objects that were deleted
		processed_classes.each do |khlass|
			smtklass = nil
			if @top_level_sorts.include? khlass 
				smtklass = khlass
			else
				if @polymorphic_parentsmap.has_key? khlass
					smtklass = "PolymorphicClass"
				else
					smtklass = @subset_table[khlass]
				end
			end
			
			@belongsto_rels[khlass].each do |rel|
				relclass = rel[0]
				relname = rel[1]
				processed_relations.push relname
				supclass = nil

				if not @top_level_sorts.include? relclass 
					if (@polymorphic_parentsmap.has_key? relclass or @polymorphic_classmap.has_key? relclass)
						relclass = "PolymorphicClass"
					else
						relclass = @subset_table[relclass]
						if @polymorphic_parentsmap.has_key? relclass or @polymorphic_classmap.has_key? relclass
							supclass = relclass
							relclass = "PolymorphicClass"
						end
					end
				end
				
				fxn_body = ""
				fxn_body << "\t(forall ((a " << smtklass << ") (b " << relclass << "))\n"
				fxn_body << "\t\t(ite (and (= b (" << relname << " a)) (Post_" << khlass << " a) "
				fxn_body << "(is" << supclass << " b) " if supclass
				fxn_body << ")\n"
				fxn_body << "\t\t    (Post_" << relname << " a b) (not (Post_" << relname << " a b))  ))\n"
				post_relations[relname] = fxn_body
				post_relation_dep[relname] = khlass
				
			end
		end
		
		#Define Post_Relation() for remaining relations
		@relations.each do |relation|
			next if processed_relations.include? relation[:relation_name]
			
			relname = relation[:relation_name]
			reltype = relation[:relation_type]
			classA = relation[:classA_name]
			classB = relation[:classB_name]
			
			if not @top_level_sorts.include? classA
				if @polymorphic_parentsmap.has_key? classA or @polymorphic_classmap.has_key? classA
					classA = "PolymorphicClass"
				else
					classA = @subset_table[classA]
					if @polymorphic_parentsmap.has_key? classA or @polymorphic_classmap.has_key? classA
						classA = "PolymorphicClass"
					end
				end
			end
			if not @top_level_sorts.include? classB
				if @polymorphic_parentsmap.has_key? classB or @polymorphic_classmap.has_key? classB
					classB = "PolymorphicClass"
				else
					classB = @subset_table[classB]
					if @polymorphic_parentsmap.has_key? classB or @polymorphic_classmap.has_key? classB
						classB = "PolymorphicClass"
					end
				end
			end

			begin
				fxn_body = "\t(forall ((b " + classA + ") (c " + classB + "))\n"								
				if reltype == RelationType::MANY_TO_MANY
					fxn_body << "\t\t  (ite (" << relname << " b c)"
				else
					fxn_body << "\t\t  (ite (= c (" << relname << " b))"
				end			
				fxn_body << " (Post_" << relname << " b c)\n"
				fxn_body << "\t\t  (not (Post_" << relname << " b c))\n  \t))\n"
				post_relations[relname] = fxn_body
			rescue 
				###TODO: Log error
			end
		end
		
		
		# Choose what to print based on projection classes and relations
		do_projection = projected_classes or projected_relations
		complete_classes = nil  # set of classes for which to print post_object fxns
						# (complete set of classes after projection completed)
		complete_relations = nil  # set of relations to print
		need_sorts_for = []	# classes to add to projected_classes for which 
						#  sort definitions are needed
		
		if do_projection					
			# Intialize 'complete_classes' and 'complete_relations'
			if projected_classes
				complete_classes = Array.new(projected_classes)
			else
				complete_classes = []
			end
			
			if projected_relations
				complete_relations = Array.new(projected_relations)
			else
				complete_relations = []
			end
			
			# Populate 'complete_classes' and 'complete_relations' 
			#  by following post_fxn definition dependencies to see which  
			#  classes and relations should be included in projected spec
			was_updated = true
			while was_updated
				was_updated = false
				class_updates = []
				relation_updates = []
				
				complete_relations.each do |r|
					if (post_relation_dep[r] and 
					  not complete_classes.include? post_relation_dep[r])
						class_updates << post_relation_dep[r]
						was_updated = true
					end
				end
				complete_classes.concat class_updates
				
				complete_classes.each do |c|
					if (post_object_depR[c] and
					  not complete_relations.include? post_object_depR[c])
						relation_updates << post_object_depR[c]
						was_updated = true
					end
				end
				complete_relations.concat relation_updates				
			end

			# Accumulate additional classes for which need sort definitions for
			to_process = complete_classes  # classes to process this round
			next_process = []  # classes to process next round (any new ones found)

			while to_process.length > 0				
				to_process.each do |c|
					dep_class = post_object_depO[c] # class that the post_object
						# definition is dependent on
					# make sure 'dep_class' part of projection set
					if (dep_class and  
					  not complete_classes.include? dep_class and
					  not need_sorts_for.include? dep_class and
					  not to_process.include? dep_class)
						next_process << dep_class
					end
				end
				need_sorts_for.concat next_process
				yo = next_process.length
				to_process = Array.new(next_process)
				next_process = []
				exit if yo != to_process.length  # sanity check
			end	
			
		else	# Print entire assert stub
			complete_classes = post_objects.keys
			complete_relations = post_relations.keys
		end
		
	
		# Create assert stub
		assert = header	# to return
		complete_classes.each do |c|
			assert << post_objects[c]
		end
		complete_relations.each do |r|
			assert << post_relations[r]
		end
		assert << footer


		if do_projection
			return [complete_classes + need_sorts_for, complete_relations, assert]
		else
			return nil, nil, assert
		end
	end
	

end #end-class


class RelationType
	ONE_TO_ONE = :one_to_one
	ONE_TO_MANY = :one_to_many
	MANY_TO_MANY  = :many_to_many
end

class ThroughRelType
	TYPE1 = 1
	TYPE2 = 2
end